import pathlib

import nose

from common import *


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo"),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_abort_stor(loop, client, server, *, tmp_dir):

    @asyncio.coroutine
    def request_abort():

        yield from asyncio.sleep(0.1, loop=loop)
        yield from client.abort()
        future.set_result(True)

    future = asyncio.Future(loop=loop)
    yield from client.login()
    stream = yield from client.upload_stream("/test.txt")
    asyncio.async(request_abort(), loop=loop)
    while True:

        try:

            for _ in range(5):

                yield from stream.write(b"-" * 8192)
                yield from asyncio.sleep(0.1, loop=loop)

        except (ConnectionResetError, BrokenPipeError):

            break

    yield from stream.finish()
    yield from future
    yield from client.quit()
    file = (tmp_dir / "test.txt")
    nose.tools.ok_(file.stat().st_size)
    file.unlink()


class FakeSlowPathIO(aioftp.PathIO):

    @asyncio.coroutine
    def exists(self, path):

        return True

    @asyncio.coroutine
    def is_file(self, path):

        return True

    @asyncio.coroutine
    def open(self, path, *args, **kwargs):

        return

    @asyncio.coroutine
    def close(self, *args, **kwargs):

        return

    @asyncio.coroutine
    def read(self, *args, **kwargs):

        yield from asyncio.sleep(1, loop=self.loop)
        return b"-" * 8192


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo"),)],
                 {"path_io_factory": FakeSlowPathIO}))
@with_connection
@with_tmp_dir("foo")
def test_abort_retr(loop, client, server, *, tmp_dir):

    @asyncio.coroutine
    def request_abort():

        yield from asyncio.sleep(0.1, loop=loop)
        yield from client.abort()
        future.set_result(True)

    abort_requested = False
    future = asyncio.Future(loop=loop)
    yield from client.login()
    stream = yield from client.download_stream("/test.txt")
    asyncio.async(request_abort(), loop=loop)
    while True:

        try:

            data = yield from stream.read(8192)
            if not data:

                break

        except (ConnectionResetError, BrokenPipeError):

            break

    yield from stream.finish()
    yield from future
    yield from client.quit()


if __name__ == "__main__":

    import logging
    import os

    os.environ["AIOFTP_TESTS"] = "PathIO"
    logging.basicConfig(
        level=logging.INFO
    )
    # this test for testing infinite loop of
    # WARNING:asyncio:socket.send() raised exception
    test_abort_stor()
    # test_abort_retr()
    print("done")



import nose

from common import *


@nose.tools.raises(aioftp.PathIsNotAbsolute)
def test_user_not_absolute_home():

    aioftp.User(home_path="foo")


def test_user_representation():

    u = aioftp.User("foo", "bar")
    nose.tools.eq_(
        repr(u),
        "User('foo', 'bar', base_path=PosixPath('.'), "
        "home_path=PurePosixPath('/'), permissions=" +
        repr([aioftp.Permission()]) +
        ", maximum_connections=" + repr(u.maximum_connections) + ")"
    )



from common import *
import aioftp


@aioftp_setup()
@with_connection
def test_multiply_connections_no_limits(loop, client, server):

    clients = [aioftp.Client(loop=loop) for _ in range(4)]
    for client in clients:

        yield from client.connect("127.0.0.1", PORT)
        yield from client.login()

    for client in clients:

        yield from client.quit()


@aioftp_setup(
    server_args=([(aioftp.User(maximum_connections=4),)], {}))
@expect_codes_in_exception("530")
@with_connection
def test_multiply_connections_limited_error(loop, client, server):

    clients = [aioftp.Client(loop=loop) for _ in range(5)]
    for client in clients:

        yield from client.connect("127.0.0.1", PORT)
        yield from client.login()

    for client in clients:

        yield from client.quit()


@aioftp_setup(
    server_args=([(aioftp.User("foo", maximum_connections=4),)], {}))
@expect_codes_in_exception("530")
@with_connection
def test_multiply_connections_with_user_limited_error(loop, client, server):

    clients = [aioftp.Client(loop=loop) for _ in range(5)]
    for client in clients:

        yield from client.connect("127.0.0.1", PORT)
        yield from client.login("foo")

    for client in clients:

        yield from client.quit()


@aioftp_setup(
    server_args=([(aioftp.User("foo", maximum_connections=4),)], {}))
@with_connection
def test_multiply_connections_relogin_balanced(loop, client, server):

    clients = [aioftp.Client(loop=loop) for _ in range(5)]
    for client in clients[:-1]:

        yield from client.connect("127.0.0.1", PORT)
        yield from client.login("foo")

    yield from clients[0].quit()
    yield from clients[-1].connect("127.0.0.1", PORT)
    yield from clients[-1].login("foo")

    for client in clients[1:]:

        yield from client.quit()


@aioftp_setup(
    server_args=([], {"maximum_connections": 5}))
@with_connection
@expect_codes_in_exception("421")
def test_multiply_connections_server_limit_error(loop, client, server):

    clients = [aioftp.Client(loop=loop) for _ in range(5)]
    for client in clients:

        yield from client.connect("127.0.0.1", PORT)
        # yield from client.login("foo")

    for client in clients:

        yield from client.quit()


@aioftp_setup(
    server_args=([], {"maximum_connections": 5}))
@with_connection
def test_multiply_connections_server_relogin_balanced(loop, client, server):

    clients = [aioftp.Client(loop=loop) for _ in range(5)]
    for client in clients[:-1]:

        yield from client.connect("127.0.0.1", PORT)

    yield from clients[0].quit()
    yield from clients[-1].connect("127.0.0.1", PORT)

    for client in clients[1:]:

        yield from client.quit()


if __name__ == "__main__":

    import logging
    import os

    os.environ["AIOFTP_TESTS"] = "PathIO"
    logging.basicConfig(
        level=logging.INFO
    )

    test_multiply_connections_server_relogin_balanced()
    print("done")



import nose
import contextlib

import aioftp.client
from common import *


@nose.tools.raises(ConnectionResetError)
@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="tests/foo", home_path="/"),)],
        {"idle_timeout": 1}))
@with_connection
def test_idle_timeout(loop, client, server):

    yield from asyncio.sleep(2, loop=loop)
    yield from client.login()


class SlowMemoryPathIO(aioftp.MemoryPathIO):

    @asyncio.coroutine
    def mkdir(self, path, parents=False):

        yield from asyncio.sleep(10, loop=self.loop)


@nose.tools.raises(ConnectionResetError)
@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="tests/foo", home_path="/"),)],
        {
            "path_timeout": 1,
            "path_io_factory": SlowMemoryPathIO,
        }))
@with_connection
def test_path_timeout(loop, client, server):

    yield from client.login()
    yield from client.make_directory("foo")


@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="tests/foo", home_path="/"),)],
        {}))
@with_connection
@expect_codes_in_exception("503")
@with_tmp_dir("foo")
def test_wait_pasv_timeout_fail_short(loop, client, server, *, tmp_dir):

    f = tmp_dir / "foo.txt"
    b = b"foobar"

    yield from client.login()
    yield from client.command("STOR " + f.name)
    yield from asyncio.sleep(0.5, loop=loop)
    reader, writer = yield from client.get_passive_connection("I")

    with contextlib.closing(writer) as writer:

        writer.write(b)
        yield from writer.drain()

    yield from client.command(None, "2xx", "1xx")
    yield from client.quit()

    with f.open("rb") as fin:

        rb = fin.read()

    f.unlink()

    nose.tools.eq_(b, rb)


@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="tests/foo", home_path="/"),)],
        {}))
@with_connection
@expect_codes_in_exception("503")
@with_tmp_dir("foo")
def test_wait_pasv_timeout_fail_long(loop, client, server, *, tmp_dir):

    f = tmp_dir / "foo.txt"
    b = b"foobar"

    yield from client.login()
    yield from client.command("STOR " + f.name)
    yield from asyncio.sleep(2, loop=loop)
    reader, writer = yield from client.get_passive_connection("I")

    with contextlib.closing(writer) as writer:

        writer.write(b)
        yield from writer.drain()

    yield from client.command(None, "2xx", "1xx")
    yield from client.quit()

    with f.open("rb") as fin:

        rb = fin.read()

    f.unlink()

    nose.tools.eq_(b, rb)


@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="tests/foo", home_path="/"),)],
        {}))
@with_connection
@with_tmp_dir("foo")
def test_wait_pasv_timeout_ok(loop, client, server, *, tmp_dir):

    f = tmp_dir / "foo.txt"
    b = b"foobar"

    yield from client.login()
    yield from client.command("TYPE I", "200")
    code, info = yield from client.command("PASV", "227")
    ip, port = client.parse_address_response(info[-1])

    yield from client.command("STOR " + f.name)
    yield from asyncio.sleep(0.5, loop=loop)
    reader, writer = yield from aioftp.client.open_connection(
        ip,
        port,
        loop,
        client.create_connection,
        read_speed_limit=client.read_speed_limit,
        read_memory=client.read_memory,
        write_speed_limit=client.write_speed_limit,
        write_memory=client.write_memory
    )

    with contextlib.closing(writer) as writer:

        writer.write(b)
        yield from writer.drain()

    yield from client.command(None, "2xx", "1xx")
    yield from client.quit()

    with f.open("rb") as fin:

        rb = fin.read()

    f.unlink()

    nose.tools.eq_(b, rb)


@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="tests/foo", home_path="/"),)],
        {}))
@with_connection
@expect_codes_in_exception("425")
@with_tmp_dir("foo")
def test_wait_pasv_timeout_ok_but_too_long(loop, client, server, *, tmp_dir):

    f = tmp_dir / "foo.txt"
    b = b"foobar"

    yield from client.login()
    yield from client.command("TYPE I", "200")
    code, info = yield from client.command("PASV", "227")
    ip, port = client.parse_address_response(info[-1])

    yield from client.command("STOR " + f.name)
    yield from asyncio.sleep(2, loop=loop)

    reader, writer = yield from aioftp.client.open_connection(
        ip,
        port,
        loop,
        client.create_connection,
        read_speed_limit=client.read_speed_limit,
        read_memory=client.read_memory,
        write_speed_limit=client.write_speed_limit,
        write_memory=client.write_memory
    )

    with contextlib.closing(writer) as writer:

        writer.write(b)
        yield from writer.drain()

    yield from client.command(None, "2xx", "1xx")
    yield from client.quit()

    with f.open("rb") as fin:

        rb = fin.read()

    f.unlink()

    nose.tools.eq_(b, rb)


if __name__ == "__main__":

    import logging
    import os

    os.environ["AIOFTP_TESTS"] = "PathIO"
    logging.basicConfig(
        level=logging.INFO
    )

    test_wait_pasv_timeout_ok()
    test_wait_pasv_timeout_ok_but_too_long()
    print("done")



import pathlib

import nose

from common import *


@aioftp_setup()
@with_connection
def test_current_directory_simple(loop, client, server):

    yield from client.login()
    cwd = yield from client.get_current_directory()
    nose.tools.eq_(cwd, pathlib.Path("/"))


@aioftp_setup(
    server_args=([(aioftp.User(home_path="/foo"),)], {}))
@with_connection
def test_current_directory_not_default(loop, client, server):

    yield from client.login()
    cwd = yield from client.get_current_directory()
    nose.tools.eq_(cwd, pathlib.Path("/foo"))


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo", home_path="/"),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_mlsd(loop, client, server, *, tmp_dir):

    tmp_file = tmp_dir / "test.txt"
    tmp_file.touch()

    yield from client.login()
    cwd = yield from client.get_current_directory()
    nose.tools.eq_(cwd, pathlib.Path("/"))

    (path, stat), *_ = files = yield from client.list()
    nose.tools.eq_(len(files), 1)

    nose.tools.eq_(path, pathlib.Path("test.txt"))
    nose.tools.eq_(stat["type"], "file")

    tmp_file.unlink()


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo", home_path="/"),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_resolving_double_dots(loop, client, server, *, tmp_dir):

    tmp_file = tmp_dir / "test.txt"
    tmp_file.touch()

    yield from client.login()

    @asyncio.coroutine
    def f():

        cwd = yield from client.get_current_directory()
        nose.tools.eq_(cwd, pathlib.Path("/"))

        (path, stat), *_ = files = yield from client.list()
        nose.tools.eq_(len(files), 1)

        nose.tools.eq_(path, pathlib.Path("test.txt"))
        nose.tools.eq_(stat["type"], "file")

    yield from f()
    yield from client.change_directory("../../../")
    yield from f()
    yield from client.change_directory("/a/../b/..")
    yield from f()

    tmp_file.unlink()



import functools
import pathlib

import nose

from common import *


def test_parse_directory_response():

    parse = functools.partial(aioftp.Client.parse_directory_response, None)

    nose.tools.eq_(
        parse('foo "baz "" test nop" """""fdfs """'),
        pathlib.Path('baz " test nop'),
    )



import pathlib
import io
import contextlib

import nose

from common import *


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo"),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_remove_single_file(loop, client, server, *, tmp_dir):

    tmp_file = tmp_dir / "foo.txt"
    nose.tools.ok_(tmp_file.exists() == False)
    tmp_file.touch()
    nose.tools.ok_(tmp_file.exists())

    yield from client.login()
    yield from client.remove_file("foo.txt")
    yield from client.quit()

    nose.tools.ok_(tmp_file.exists() == False)


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo"),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_recursive_remove(loop, client, server, *, tmp_dir):

    d = tmp_dir / "foo"
    d.mkdir()
    (d / "bar.txt").touch()
    (d / "baz.foo").touch()
    dd = d / "bar_dir"
    dd.mkdir()
    (dd / "foo.baz").touch()

    yield from client.login()
    yield from client.remove("foo")
    r = yield from client.list()
    yield from client.quit()

    nose.tools.eq_(len(r), 0)


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo"),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_file_download(loop, client, server, *, tmp_dir):

    f = tmp_dir / "foo"
    with f.open("w") as fout:

        fout.write("foobar")

    yield from client.login()
    stream = yield from client.download_stream("foo")
    data = yield from stream.read()
    yield from stream.finish()
    yield from client.quit()

    f.unlink()

    nose.tools.eq_(data, b"foobar")


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo"),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_file_upload(loop, client, server, *, tmp_dir):

    def callback(data):

        nose.tools.eq_(data, b)

    f = tmp_dir / "foo"
    b = b"foobar"

    yield from client.login()

    stream = yield from client.upload_stream("foo")
    yield from stream.write(b)
    yield from stream.finish()
    yield from client.quit()

    with f.open("rb") as fin:

        rb = fin.read()

    f.unlink()

    nose.tools.eq_(b, rb)


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo"),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_file_append(loop, client, server, *, tmp_dir):

    f = tmp_dir / "foo"
    with f.open("w") as fout:

        fout.write("foobar")

    ab = b"foobar"
    file_like = io.BytesIO(ab)
    yield from client.login()
    stream = yield from client.append_stream("foo")
    yield from stream.write(ab)
    yield from stream.finish()
    yield from client.quit()

    with f.open("rb") as fin:

        rb = fin.read()

    f.unlink()

    nose.tools.eq_(b"foobar" + ab, rb)


@contextlib.contextmanager
def make_some_files(path):

    for i in range(4):

        f = path / str.format("file{}", i)
        with f.open("w") as fout:

            fout.write("foobar")

    yield
    for i in range(4):

        f = path / str.format("file{}", i)
        f.unlink()


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo/server"),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_upload_folder(loop, client, server, *, tmp_dir):

    sdir = tmp_dir / "server"
    sdir.mkdir()

    cdir = tmp_dir / "client"
    cdir.mkdir()

    ecdir = cdir / "extra"
    ecdir.mkdir()

    with make_some_files(cdir), make_some_files(ecdir):

        spaths = set(
            map(
                lambda p: p.relative_to(cdir),
                cdir.rglob("*")
            )
        )
        yield from client.login()
        yield from client.upload(cdir)
        rpaths = set(
            map(
                lambda p: p.relative_to(sdir / "client"),
                (sdir / "client").rglob("*"),
            )
        )
        yield from client.remove("/")
        yield from client.quit()

    ecdir.rmdir()
    cdir.rmdir()
    nose.tools.eq_(spaths, rpaths)


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo/server"),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_upload_folder_into(loop, client, server, *, tmp_dir):

    sdir = tmp_dir / "server"
    sdir.mkdir()

    cdir = tmp_dir / "client"
    cdir.mkdir()

    ecdir = cdir / "extra"
    ecdir.mkdir()

    with make_some_files(cdir), make_some_files(ecdir):

        spaths = set(
            map(
                lambda p: p.relative_to(cdir),
                cdir.rglob("*")
            )
        )
        yield from client.login()
        yield from client.upload(cdir, write_into=True)
        rpaths = set(
            map(
                lambda p: p.relative_to(sdir),
                (sdir).rglob("*"),
            )
        )
        yield from client.remove("/")
        yield from client.quit()

    ecdir.rmdir()
    cdir.rmdir()
    nose.tools.eq_(spaths, rpaths)


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo/server"),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_upload_folder_into_another(loop, client, server, *, tmp_dir):

    sdir = tmp_dir / "server"
    sdir.mkdir()

    esdir = sdir / "foo"
    esdir.mkdir()

    cdir = tmp_dir / "client"
    cdir.mkdir()

    ecdir = cdir / "extra"
    ecdir.mkdir()

    with make_some_files(cdir), make_some_files(ecdir):

        spaths = set(
            map(
                lambda p: p.relative_to(cdir),
                cdir.rglob("*")
            )
        )
        yield from client.login()
        yield from client.upload(cdir, "foo", write_into=True)
        rpaths = set(
            map(
                lambda p: p.relative_to(esdir),
                (esdir).rglob("*"),
            )
        )
        yield from client.remove("/")
        yield from client.quit()

    ecdir.rmdir()
    cdir.rmdir()
    nose.tools.eq_(spaths, rpaths)


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo"),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_download_folder(loop, client, server, *, tmp_dir):

    sdir = tmp_dir / "server"
    sdir.mkdir()

    esdir = sdir / "extra"
    esdir.mkdir()

    cdir = tmp_dir / "client"
    cdir.mkdir()

    with make_some_files(sdir), make_some_files(esdir):

        spaths = set(
            map(
                lambda p: p.relative_to(sdir),
                sdir.rglob("*")
            )
        )
        yield from client.login()
        yield from client.download("/server", cdir)
        cpaths = set(
            map(
                lambda p: p.relative_to(cdir / "server"),
                (cdir / "server").rglob("*"),
            )
        )
        yield from client.remove("/client")
        yield from client.quit()

    esdir.rmdir()
    sdir.rmdir()
    nose.tools.eq_(spaths, cpaths)


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo"),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_download_folder_into(loop, client, server, *, tmp_dir):

    sdir = tmp_dir / "server"
    sdir.mkdir()

    esdir = sdir / "extra"
    esdir.mkdir()

    cdir = tmp_dir / "client"
    cdir.mkdir()

    with make_some_files(sdir), make_some_files(esdir):

        spaths = set(
            map(
                lambda p: p.relative_to(sdir),
                sdir.rglob("*")
            )
        )
        yield from client.login()
        yield from client.download("/server", cdir, write_into=True)
        cpaths = set(
            map(
                lambda p: p.relative_to(cdir),
                (cdir).rglob("*"),
            )
        )
        yield from client.remove("/client")
        yield from client.quit()

    esdir.rmdir()
    sdir.rmdir()
    nose.tools.eq_(spaths, cpaths)


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo"),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_download_folder_into_another(loop, client, server, *, tmp_dir):

    sdir = tmp_dir / "server"
    sdir.mkdir()

    esdir = sdir / "extra"
    esdir.mkdir()

    cdir = tmp_dir / "client"
    cdir.mkdir()

    with make_some_files(sdir), make_some_files(esdir):

        spaths = set(
            map(
                lambda p: p.relative_to(sdir),
                sdir.rglob("*")
            )
        )
        yield from client.login()
        yield from client.download("/server", cdir / "foo", write_into=True)
        cpaths = set(
            map(
                lambda p: p.relative_to(cdir / "foo"),
                (cdir / "foo").rglob("*"),
            )
        )
        yield from client.remove("/client")
        yield from client.quit()

    esdir.rmdir()
    sdir.rmdir()
    nose.tools.eq_(spaths, cpaths)


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo"),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_upload_file_into_another(loop, client, server, *, tmp_dir):

    cfile = tmp_dir / "client_file.txt"
    sfile = tmp_dir / "server_file.txt"
    with cfile.open("w") as fout:

        fout.write("foobar")

    yield from client.login()
    yield from client.upload(cfile, sfile.name, write_into=True)

    with sfile.open() as fin:

        data = fin.read()

    cfile.unlink()
    sfile.unlink()
    nose.tools.eq_("foobar", data)


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo"),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_download_file(loop, client, server, *, tmp_dir):

    cfile = tmp_dir / "bar" / "server_file.txt"
    sfile = tmp_dir / "server_file.txt"

    with sfile.open("w") as fout:

        fout.write("foobar")

    yield from client.login()
    yield from client.download(sfile.name, tmp_dir / "bar")

    with cfile.open() as fin:

        data = fin.read()

    cfile.unlink()
    cfile.parent.rmdir()
    sfile.unlink()
    nose.tools.eq_("foobar", data)


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo"),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_download_file_write_into(loop, client, server, *, tmp_dir):

    cfile = tmp_dir / "client_file.txt"
    sfile = tmp_dir / "server_file.txt"

    with sfile.open("w") as fout:

        fout.write("foobar")

    yield from client.login()
    yield from client.download(sfile.name, cfile, write_into=True)

    with cfile.open() as fin:

        data = fin.read()

    cfile.unlink()
    sfile.unlink()
    nose.tools.eq_("foobar", data)


if __name__ == "__main__":

    import logging
    import os

    os.environ["AIOFTP_TESTS"] = "PathIO"
    logging.basicConfig(
        level=logging.INFO
    )
    # test_remove_single_file()
    # test_recursive_remove()
    # test_file_download()
    # test_file_upload()
    # test_file_append()
    test_upload_folder()
    test_upload_folder_into()
    test_upload_folder_into_another()
    test_download_folder()
    test_download_folder_into()
    test_download_folder_into_another()
    test_upload_file_into_another()
    test_download_file()
    test_download_file_write_into()
    print("done")



import asyncio
import functools
import collections
import operator
import io
import time


__all__ = (
    "AbstractPathIO",
    "PathIO",
    "AsyncPathIO",
    "MemoryPathIO",
)


class AbstractPathIO:
    """
    Abstract class for path io operations.

    :param loop: loop to use
    :type loop: :py:class:`asyncio.BaseEventLoop`
    """

    def __init__(self, loop=None):

        self.loop = loop or asyncio.get_event_loop()

    @asyncio.coroutine
    def exists(self, path):
        """
        :py:func:`asyncio.coroutine`

        Check if path exists

        :param path: path to check
        :type path: :py:class:`pathlib.Path`

        :rtype: :py:class:`bool`
        """
        raise NotImplementedError

    @asyncio.coroutine
    def is_dir(self, path):
        """
        :py:func:`asyncio.coroutine`

        Check if path is directory

        :param path: path to check
        :type path: :py:class:`pathlib.Path`

        :rtype: :py:class:`bool`
        """
        raise NotImplementedError

    @asyncio.coroutine
    def is_file(self, path):
        """
        :py:func:`asyncio.coroutine`

        Check if path is file

        :param path: path to check
        :type path: :py:class:`pathlib.Path`

        :rtype: :py:class:`bool`
        """
        raise NotImplementedError

    @asyncio.coroutine
    def mkdir(self, path, *, parents=False):
        """
        :py:func:`asyncio.coroutine`

        Make directory

        :param path: path to create
        :type path: :py:class:`pathlib.Path`

        :param parents: create parents is does not exists
        :type parents: :py:class:`bool`
        """
        raise NotImplementedError

    @asyncio.coroutine
    def rmdir(self, path):
        """
        :py:func:`asyncio.coroutine`

        Remove directory

        :param path: path to remove
        :type path: :py:class:`pathlib.Path`
        """
        raise NotImplementedError

    @asyncio.coroutine
    def unlink(self, path):
        """
        :py:func:`asyncio.coroutine`

        Remove file

        :param path: path to remove
        :type path: :py:class:`pathlib.Path`
        """
        raise NotImplementedError

    @asyncio.coroutine
    def list(self, path):
        """
        :py:func:`asyncio.coroutine`

        List path content. If path is file, then return empty sequence

        :param path: path to list
        :type path: :py:class:`pathlib.Path`

        :rtype: :py:class:`collections.Iterable` of :py:class:`pathlib.Path`
        """
        raise NotImplementedError

    @asyncio.coroutine
    def stat(self, path):
        """
        :py:func:`asyncio.coroutine`

        Get path stats

        :param path: path, which stats need
        :type path: :py:class:`pathlib.Path`

        :return: path stats. For proper work you need only this stats:
          st_size, st_mtime, st_ctime, st_nlink, st_mode
        :rtype: same as :py:class:`os.stat_result`
        """
        raise NotImplementedError

    @asyncio.coroutine
    def open(self, path, *args, **kwargs):
        """
        :py:func:`asyncio.coroutine`

        Open file. You should implement "mode" argument, which can be:
        "rb", "wb", "ab" (read, write, append. all binary). Return type depends
        on implementation, anyway the only place you need this file-object
        is in your implementation of read, write and close

        :param path: path to create
        :type path: :py:class:`pathlib.Path`

        :return: file-object
        """
        raise NotImplementedError

    @asyncio.coroutine
    def write(self, file, data):
        """
        :py:func:`asyncio.coroutine`

        Write some data to file

        :param file: file-object from :py:class:`aioftp.AbstractPathIO.open`

        :param data: data to write
        :type data: :py:class:`bytes`
        """
        raise NotImplementedError

    @asyncio.coroutine
    def read(self, file):
        """
        :py:func:`asyncio.coroutine`

        Read some data from file

        :param file: file-object from :py:class:`aioftp.AbstractPathIO.open`

        :rtype: :py:class:`bytes`
        """
        raise NotImplementedError

    @asyncio.coroutine
    def close(self, file):
        """
        :py:func:`asyncio.coroutine`

        Close file

        :param file: file-object from :py:class:`aioftp.AbstractPathIO.open`
        """
        raise NotImplementedError

    @asyncio.coroutine
    def rename(self, source, destination):
        """
        :py:func:`asyncio.coroutine`

        Rename path

        :param source: rename from
        :type source: :py:class:`pathlib.Path`

        :param destination: rename to
        :type destination: :py:class:`pathlib.Path`
        """
        raise NotImplementedError


class PathIO(AbstractPathIO):
    """
    Blocking path io. Directly based on :py:class:`pathlib.Path` methods.
    """

    @asyncio.coroutine
    def exists(self, path):

        return path.exists()

    @asyncio.coroutine
    def is_dir(self, path):

        return path.is_dir()

    @asyncio.coroutine
    def is_file(self, path):

        return path.is_file()

    @asyncio.coroutine
    def mkdir(self, path, *, parents=False):

        return path.mkdir(parents=parents)

    @asyncio.coroutine
    def rmdir(self, path):

        return path.rmdir()

    @asyncio.coroutine
    def unlink(self, path):

        return path.unlink()

    @asyncio.coroutine
    def list(self, path):

        return tuple(path.glob("*"))

    @asyncio.coroutine
    def stat(self, path):

        return path.stat()

    @asyncio.coroutine
    def open(self, path, *args, **kwargs):

        return path.open(*args, **kwargs)

    @asyncio.coroutine
    def write(self, file, data):

        return file.write(data)

    @asyncio.coroutine
    def read(self, file, *args, **kwargs):

        return file.read(*args, **kwargs)

    @asyncio.coroutine
    def close(self, file):

        return file.close()

    @asyncio.coroutine
    def rename(self, source, destination):

        return source.rename(destination)


class AsyncPathIO(AbstractPathIO):
    """
    Non-blocking path io. Based on
    :py:meth:`asyncio.BaseEventLoop.run_in_executor` and
    :py:class:`pathlib.Path` methods.
    """

    @asyncio.coroutine
    def exists(self, path):

        return (yield from self.loop.run_in_executor(None, path.exists))

    @asyncio.coroutine
    def is_dir(self, path):

        return (yield from self.loop.run_in_executor(None, path.is_dir))

    @asyncio.coroutine
    def is_file(self, path):

        return (yield from self.loop.run_in_executor(None, path.is_file))

    @asyncio.coroutine
    def mkdir(self, path, *, parents=False):

        f = functools.partial(path.mkdir, parents=parents)
        return (yield from self.loop.run_in_executor(None, f))

    @asyncio.coroutine
    def rmdir(self, path):

        return (yield from self.loop.run_in_executor(None, path.rmdir))

    @asyncio.coroutine
    def unlink(self, path):

        return (yield from self.loop.run_in_executor(None, path.unlink))

    @asyncio.coroutine
    def list(self, path):

        def worker(pattern):

            return tuple(path.glob(pattern))

        return (yield from self.loop.run_in_executor(None, worker, "*"))

    @asyncio.coroutine
    def stat(self, path):

        return (yield from self.loop.run_in_executor(None, path.stat))

    @asyncio.coroutine
    def open(self, path, *args, **kwargs):

        f = functools.partial(path.open, *args, **kwargs)
        return (yield from self.loop.run_in_executor(None, f))

    @asyncio.coroutine
    def write(self, file, data):

        return (yield from self.loop.run_in_executor(None, file.write, data))

    @asyncio.coroutine
    def read(self, file, *args, **kwargs):

        f = functools.partial(file.read, *args, **kwargs)
        return (yield from self.loop.run_in_executor(None, f))

    @asyncio.coroutine
    def close(self, file):

        return (yield from self.loop.run_in_executor(None, file.close))

    @asyncio.coroutine
    def rename(self, source, destination):

        f = functools.partial(source.rename, destination)
        return (yield from self.loop.run_in_executor(None, f))


class Node:

    def __init__(self, type, name, ctime=None, mtime=None, *, content):

        self.type = type
        self.name = name
        self.ctime = ctime or int(time.time())
        self.mtime = mtime or int(time.time())
        self.content = content

    def __repr__(self):

        return str.format(
            "{}(type={!r}, name={!r}, ctime={!r}, mtime={!r}, content={!r})",
            self.__class__.__name__,
            self.type,
            self.name,
            self.ctime,
            self.mtime,
            self.content,
        )


class MemoryPathIO(AbstractPathIO):
    """
    Non-blocking path io. Based on in-memory tree. It is just proof of concept
    and probably not so fast as it can be.
    """

    Stats = collections.namedtuple(
        "Stats",
        (
            "st_size",
            "st_ctime",
            "st_mtime",
            "st_nlink",
            "st_mode",
        )
    )

    def __init__(self, loop=None):

        super().__init__(loop)
        self.fs = [Node("dir", "/", content=[])]

    def __repr__(self):

        return repr(self.fs)

    def get_node(self, path):

        nodes = self.fs
        for part in path.parts:

            if not isinstance(nodes, list):

                return

            for node in nodes:

                if node.name == part:

                    nodes = node.content
                    break

            else:

                return

        return node

    @asyncio.coroutine
    def exists(self, path):

        return self.get_node(path) is not None

    @asyncio.coroutine
    def is_dir(self, path):

        node = self.get_node(path)
        return not (node is None or node.type != "dir")

    @asyncio.coroutine
    def is_file(self, path):

        node = self.get_node(path)
        return not (node is None or node.type != "file")

    @asyncio.coroutine
    def mkdir(self, path, *, parents=False):

        if self.get_node(path):

            raise FileExistsError

        elif not parents:

            parent = self.get_node(path.parent)
            if parent is None:

                raise FileNotFoundError

            elif parent.type != "dir":

                raise FileExistsError

            node = Node("dir", path.name, content=[])
            parent.content.append(node)

        else:

            nodes = self.fs
            for part in path.parts:

                if isinstance(nodes, list):

                    for node in nodes:

                        if node.name == part:

                            nodes = node.content
                            break

                    else:

                        node = Node("dir", part, content=[])
                        nodes.append(node)
                        nodes = node.content

                else:

                    raise FileExistsError

    @asyncio.coroutine
    def rmdir(self, path):

        node = self.get_node(path)
        if node is None:

            raise FileNotFoundError

        elif node.type != "dir":

            raise NotADirectoryError

        elif node.content:

            raise OSError("Directory not empty")

        else:

            parent = self.get_node(path.parent)
            for i, node in enumerate(parent.content):

                if node.name == path.name:

                    break

            parent.content.pop(i)

    @asyncio.coroutine
    def unlink(self, path):

        node = self.get_node(path)
        if node is None:

            raise FileNotFoundError

        elif node.type != "file":

            raise IsADirectoryError

        else:

            parent = self.get_node(path.parent)
            for i, node in enumerate(parent.content):

                if node.name == path.name:

                    break

            parent.content.pop(i)

    @asyncio.coroutine
    def list(self, path):

        node = self.get_node(path)
        if node is None or node.type != "dir":

            return ()

        else:

            names = map(operator.attrgetter("name"), node.content)
            paths = map(lambda name: path / name, names)
            return tuple(paths)

    @asyncio.coroutine
    def stat(self, path):

        node = self.get_node(path)
        if node is None:

            raise FileNotFoundError

        else:

            if node.type == "file":

                size = len(node.content.getbuffer())

            else:

                size = 0

            return MemoryPathIO.Stats(
                size,
                node.ctime,
                node.mtime,
                1,
                0o100777,
            )

    @asyncio.coroutine
    def open(self, path, mode="rb", *args, **kwargs):

        if mode == "rb":

            node = self.get_node(path)
            if node is None:

                raise FileNotFoundError

            file_like = node.content
            file_like.seek(0, io.SEEK_SET)

        elif mode in ("wb", "ab"):

            node = self.get_node(path)
            if node is None:

                parent = self.get_node(path.parent)
                if parent is None or parent.type != "dir":

                    raise FileNotFoundError

                new_node = Node("file", path.name, content=io.BytesIO())
                parent.content.append(new_node)
                file_like = new_node.content

            elif node.type != "file":

                raise IsADirectoryError

            else:

                if mode == "wb":

                    file_like = node.content = io.BytesIO()

                else:

                    file_like = node.content
                    file_like.seek(0, io.SEEK_END)

        else:

            raise ValueError(str.format("invalid mode: {}", mode))

        return file_like

    @asyncio.coroutine
    def write(self, file, data):

        file.write(data)
        file.mtime = int(time.time())

    @asyncio.coroutine
    def read(self, file, count=None):

        return file.read(count)

    @asyncio.coroutine
    def close(self, file):

        pass

    @asyncio.coroutine
    def rename(self, source, destination):

        if source != destination:

            sparent = self.get_node(source.parent)
            dparent = self.get_node(destination.parent)
            snode = self.get_node(source)
            if None in (snode, dparent):

                raise FileNotFoundError

            for i, node in enumerate(sparent.content):

                if node.name == source.name:

                    sparent.content.pop(i)

            snode.name = destination.name
            for i, node in enumerate(dparent.content):

                if node.name == destination.name:

                    dparent.content[i] = snode
                    break

            else:

                dparent.content.append(snode)



import pathlib

import nose

from common import *


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo"),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_create_and_remove_directory(loop, client, server, *, tmp_dir):

    yield from client.login()
    yield from client.make_directory("bar")
    (path, stat), *_ = files = yield from client.list()
    nose.tools.eq_(len(files), 1)
    nose.tools.eq_(path, pathlib.Path("bar"))
    nose.tools.eq_(stat["type"], "dir")

    yield from client.remove_directory("bar")
    files = yield from client.list()
    nose.tools.eq_(len(files), 0)


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo"),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_create_and_remove_directory_long(loop, client, server, *, tmp_dir):

    yield from client.login()
    yield from client.make_directory("bar/baz")
    (path, stat), *_ = files = yield from client.list()
    nose.tools.eq_(len(files), 1)
    nose.tools.eq_(path, pathlib.Path("bar"))
    nose.tools.eq_(stat["type"], "dir")

    (path, stat), *_ = files = yield from client.list("bar")
    nose.tools.eq_(len(files), 1)
    nose.tools.eq_(path, pathlib.Path("bar/baz"))
    nose.tools.eq_(stat["type"], "dir")

    yield from client.remove_directory("bar/baz")
    yield from client.remove_directory("bar")
    files = yield from client.list()
    nose.tools.eq_(len(files), 0)


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo"),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_create_directory_long_no_parents(loop, client, server, *, tmp_dir):

    yield from client.login()
    yield from client.make_directory("bar/baz", parents=False)
    yield from client.remove_directory("bar/baz")
    yield from client.remove_directory("bar")


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo"),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_change_directory(loop, client, server, *, tmp_dir):

    yield from client.login()
    yield from client.make_directory("bar")
    yield from client.change_directory("bar")
    cwd = yield from client.get_current_directory()
    nose.tools.eq_(cwd, pathlib.Path("/bar"))

    yield from client.change_directory()
    cwd = yield from client.get_current_directory()
    nose.tools.eq_(cwd, pathlib.Path("/"))

    yield from client.remove_directory("bar")
    files = yield from client.list()
    nose.tools.eq_(len(files), 0)


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo"),)], {}))
@with_connection
@expect_codes_in_exception("550")
@with_tmp_dir("foo")
def test_change_directory_not_exist(loop, client, server, *, tmp_dir):

    yield from client.login()
    yield from client.change_directory("bar")


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo"),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_rename_empty_directory(loop, client, server, *, tmp_dir):

    yield from client.login()
    yield from client.make_directory("bar")
    (path, stat), *_ = files = yield from client.list()
    nose.tools.eq_(len(files), 1)
    nose.tools.eq_(path, pathlib.Path("bar"))
    nose.tools.eq_(stat["type"], "dir")

    yield from client.rename("bar", "baz")
    (path, stat), *_ = files = yield from client.list()
    nose.tools.eq_(len(files), 1)
    nose.tools.eq_(path, pathlib.Path("baz"))
    nose.tools.eq_(stat["type"], "dir")

    yield from client.remove_directory("baz")
    files = yield from client.list()
    nose.tools.eq_(len(files), 0)


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo"),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_rename_non_empty_directory(loop, client, server, *, tmp_dir):

    yield from client.login()
    yield from client.make_directory("bar")
    (path, stat), *_ = files = yield from client.list()
    nose.tools.eq_(len(files), 1)
    nose.tools.eq_(path, pathlib.Path("bar"))
    nose.tools.eq_(stat["type"], "dir")

    tmp_file = tmp_dir / "bar" / "foo.txt"
    tmp_file.touch()

    yield from client.make_directory("hurr")
    yield from client.rename("bar", "hurr/baz")
    (path, stat), *_ = files = yield from client.list()
    nose.tools.eq_(len(files), 1)
    nose.tools.eq_(path, pathlib.Path("hurr"))
    nose.tools.eq_(stat["type"], "dir")

    (path, stat), *_ = files = yield from client.list("hurr")
    nose.tools.eq_(len(files), 1)
    nose.tools.eq_(path, pathlib.Path("hurr/baz"))
    nose.tools.eq_(stat["type"], "dir")

    (path, stat), *_ = files = yield from client.list("/hurr/baz")
    nose.tools.eq_(len(files), 1)
    nose.tools.eq_(path, pathlib.Path("/hurr/baz/foo.txt"))
    nose.tools.eq_(stat["type"], "file")

    tmp_file = tmp_dir / "hurr" / "baz" / "foo.txt"
    tmp_file.unlink()
    yield from client.remove_directory("hurr/baz")
    yield from client.remove_directory("hurr")
    files = yield from client.list()
    nose.tools.eq_(len(files), 0)



import asyncio
import functools
import sys
import pathlib
import os
import aioftp

import nose


PORT = 8888


@nose.tools.nottest
def aioftp_setup(*, server_args=([], {}), client_args=([], {})):

    def decorator(f):

        @functools.wraps(f)
        def wrapper():

            s_args, s_kwargs = server_args
            c_args, c_kwargs = client_args

            def run_in_loop(s_args, s_kwargs, c_args, c_kwargs):

                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(None)

                server = aioftp.Server(*s_args, loop=loop, **s_kwargs)
                client = aioftp.Client(*c_args, loop=loop, **c_kwargs)

                coro = asyncio.coroutine(f)
                try:

                    loop.run_until_complete(coro(loop, client, server))

                finally:

                    if hasattr(server, "server"):

                        server.close()
                        loop.run_until_complete(server.wait_closed())

                    if hasattr(client, "writer"):

                        client.close()

                    loop.close()

            if "path_io_factory" not in s_kwargs:
                for factory in (aioftp.PathIO, aioftp.AsyncPathIO):
                    s_kwargs["path_io_factory"] = factory
                    run_in_loop(s_args, s_kwargs, c_args, c_kwargs)
            else:
                run_in_loop(s_args, s_kwargs, c_args, c_kwargs)

        return wrapper

    return decorator


@nose.tools.nottest
def with_connection(f):

    @functools.wraps(f)
    def wrapper(loop, client, server):

        try:

            yield from server.start(None, PORT)
            yield from client.connect("127.0.0.1", PORT)
            yield from asyncio.coroutine(f)(loop, client, server)

        finally:

            client.close()
            server.close()
            yield from server.wait_closed()

    return wrapper


@nose.tools.nottest
def expect_codes_in_exception(*codes):

    def decorator(f):

        @functools.wraps(f)
        def wrapper(*args):

            try:

                yield from f(*args)

            except aioftp.StatusCodeError as exc:

                nose.tools.eq_(exc.received_codes, codes)

            else:

                raise Exception("There was no exception")

        return wrapper

    return decorator


@nose.tools.nottest
def with_tmp_dir(name):

    def decorator(f):

        @functools.wraps(f)
        def wrapper(*args):

            tmp_dir = pathlib.Path("tests") / name
            tmp_dir.mkdir(parents=True)
            try:

                yield from f(*args, tmp_dir=tmp_dir)

            finally:

                tmp_dir.rmdir()

        return wrapper

    return decorator



import asyncio
import re
import contextlib
import collections
import pathlib

from . import errors
from . import common


__all__ = ("Client", "Code", "StreamIO")


def add_prefix(message):

    return str.format("aioftp client: {}", message)


@asyncio.coroutine
def open_connection(host, port, loop, create_connection, *,
                    read_speed_limit, read_memory,
                    write_speed_limit, write_memory):

    reader = asyncio.StreamReader(loop=loop)
    protocol = asyncio.StreamReaderProtocol(reader, loop=loop)
    transport, _ = yield from create_connection(lambda: protocol, host, port)
    writer = asyncio.StreamWriter(transport, protocol, reader, loop)

    throttle_reader = common.Throttle(
        reader,
        loop=loop,
        throttle=read_speed_limit,
        memory=read_memory
    )
    throttle_writer = common.Throttle(
        writer,
        loop=loop,
        throttle=write_speed_limit,
        memory=write_memory
    )
    return throttle_reader, throttle_writer


class Code(str):
    """
    Representation of server status code.
    """

    def matches(self, mask):
        """
        :param mask: Template for comparision. If mask symbol is not digit
            then it passes.
        :type mask: :py:class:`str`

        ::

            >>> Code("123").matches("1")
            True
            >>> Code("123").matches("1x3")
            True
        """
        return all(map(lambda m, c: not str.isdigit(m) or m == c, mask, self))


class StreamIO:
    """
    Stream input/output wrapper.

    :param client: aioftp client
    :type client: :py:class:`aioftp.Client`

    :param reader: stream reader
    :type reader: :py:class:`asyncio.StreamReader`

    :param writer: stream writer
    :type writer: :py:class:`asyncio.StreamWriter`
    """

    def __init__(self, client, reader, writer):

        self.client = client
        self.reader = reader
        self.writer = writer

    @asyncio.coroutine
    def readline(self):
        """
        :py:func:`asyncio.coroutine`

        Proxy for :py:meth:`asyncio.StreamReader.readline`.
        """
        return (yield from self.reader.readline())

    @asyncio.coroutine
    def read(self, count=common.default_block_size):
        """
        :py:func:`asyncio.coroutine`

        Proxy for :py:meth:`asyncio.StreamReader.read`.

        :param count: block size for read operation
        :type count: :py:class:`int`
        """
        return (yield from self.reader.read(count))

    @asyncio.coroutine
    def write(self, data):
        """
        :py:func:`asyncio.coroutine`

        Combination of :py:meth:`asyncio.StreamWriter.write` and
        :py:meth:`asyncio.StreamWriter.drain`.

        :param data: data to write
        :type data: :py:class:`bytes`
        """
        self.writer.write(data)
        yield from self.writer.drain()

    @asyncio.coroutine
    def finish(self, expected_codes="2xx", wait_codes="1xx"):
        """
        :py:func:`asyncio.coroutine`

        Close connection and wait for `expected_codes` response from server
        passing `wait_codes`.

        :param expected_codes: tuple of expected codes or expected code
        :type expected_codes: :py:class:`tuple` of :py:class:`str` or
            :py:class:`str`

        :param wait_codes: tuple of wait codes or wait code
        :type wait_codes: :py:class:`tuple` of :py:class:`str` or
            :py:class:`str`
        """
        self.close()
        yield from self.client.command(None, expected_codes, wait_codes)

    def close(self):
        """
        Close connection.
        """
        self.writer.close()


class BaseClient:

    def __init__(self, *, loop=None, create_connection=None, timeout=None,
                 read_speed_limit=None, write_speed_limit=None):

        self.loop = loop or asyncio.get_event_loop()
        self.create_connection = create_connection or \
            self.loop.create_connection
        self.timeout = timeout

        self.read_speed_limit = read_speed_limit
        self.write_speed_limit = write_speed_limit

        self.read_memory = common.ThrottleMemory()
        self.write_memory = common.ThrottleMemory()

    @asyncio.coroutine
    def connect(self, host, port=21):

        self.reader, self.writer = yield from open_connection(
            host,
            port,
            self.loop,
            self.create_connection,
            read_speed_limit=self.read_speed_limit,
            read_memory=self.read_memory,
            write_speed_limit=self.write_speed_limit,
            write_memory=self.write_memory
        )

    def close(self):
        """
        Close connection.
        """
        self.writer.close()

    @asyncio.coroutine
    def parse_line(self):
        """
        :py:func:`asyncio.coroutine`

        Parsing server response line.

        :return: (code, line)
        :rtype: (:py:class:`aioftp.Code`, :py:class:`str`)

        :raises ConnectionResetError: if received data is empty (this
            means, that connection is closed)
        :raises asyncio.TimeoutError: if there where no data for `timeout`
            period
        """
        line = yield from asyncio.wait_for(
            self.reader.readline(),
            self.timeout,
            loop=self.loop,
        )
        if not line:

            self.writer.close()
            raise ConnectionResetError

        s = str.rstrip(bytes.decode(line, encoding="utf-8"))
        common.logger.info(add_prefix(s))
        return Code(s[:3]), s[3:]

    @asyncio.coroutine
    def parse_response(self):
        """
        :py:func:`asyncio.coroutine`

        Parsing full server response (all lines).

        :return: (code, lines)
        :rtype: (:py:class:`aioftp.Code`, :py:class:`list` of :py:class:`str`)

        :raises aioftp.StatusCodeError: if received code does not matches all
            already received codes
        """
        code, rest = yield from self.parse_line()
        info = [rest]
        curr_code = code
        while str.startswith(rest, "-") or not str.isdigit(curr_code):

            curr_code, rest = yield from self.parse_line()
            if str.isdigit(curr_code):

                info.append(rest)
                if curr_code != code:

                    raise errors.StatusCodeError(code, curr_code, info)

            else:

                info.append(curr_code + rest)

        return code, info

    def check_codes(self, expected_codes, received_code, info):
        """
        Checks if any of expected matches received.

        :param expected_codes: tuple of expected codes
        :type expected_codes: :py:class:`tuple`

        :param received_code: received code for matching
        :type received_code: :py:class:`aioftp.Code`

        :param info: list of response lines from server
        :type info: :py:class:`list`

        :raises aioftp.StatusCodeError: if received code does not matches any
            expected code
        """
        if not any(map(received_code.matches, expected_codes)):

            raise errors.StatusCodeError(expected_codes, received_code, info)

    @asyncio.coroutine
    def command(self, command=None, expected_codes=(), wait_codes=()):
        """
        :py:func:`asyncio.coroutine`

        Basic command logic.

        1. Send command if not omitted.
        2. Yield response until no wait code matches.
        3. Check code for expected.

        :param command: command line
        :type command: :py:class:`str`

        :param expected_codes: tuple of expected codes or expected code
        :type expected_codes: :py:class:`tuple` of :py:class:`str` or
            :py:class:`str`

        :param wait_codes: tuple of wait codes or wait code
        :type wait_codes: :py:class:`tuple` of :py:class:`str` or
            :py:class:`str`
        """
        expected_codes = common.wrap_with_container(expected_codes)
        wait_codes = common.wrap_with_container(wait_codes)

        if command:

            common.logger.info(add_prefix(command))
            message = command + common.end_of_line
            self.writer.write(str.encode(message, encoding="utf-8"))
            yield from self.writer.drain()

        if expected_codes or wait_codes:

            code, info = yield from self.parse_response()
            while any(map(code.matches, wait_codes)):

                code, info = yield from self.parse_response()

            if expected_codes:

                self.check_codes(expected_codes, code, info)

            return code, info

    def parse_address_response(self, s):
        """
        Parsing ip:port server response.

        :param s: response line
        :type s: :py:class:`str`

        :return: (ip, port)
        :rtype: (:py:class:`str`, :py:class:`int`)
        """
        sub, *_ = re.findall(r"[^(]*\(([^)]*)", s)
        nums = tuple(map(int, str.split(sub, ",")))
        ip = str.join(".", map(str, nums[:4]))
        port = (nums[4] << 8) | nums[5]
        return ip, port

    def parse_directory_response(self, s):
        """
        Parsing directory server response.

        :param s: response line
        :type s: :py:class:`str`

        :rtype: :py:class:`pathlib.Path`
        """
        seq_quotes = 0
        start = False
        directory = ""
        for ch in s:

            if not start:

                if ch == '"':

                    start = True

            else:

                if ch == '"':

                    seq_quotes += 1

                else:

                    if seq_quotes == 1:

                        break

                    elif seq_quotes == 2:

                        seq_quotes = 0
                        directory += '"'

                    directory += ch

        return pathlib.Path(directory)

    def parse_mlsx_line(self, b):
        """
        Parsing MLS(T|D) response.

        :param b: response line
        :type b: :py:class:`bytes` or :py:class:`str`

        :return: (path, info)
        :rtype: (:py:class:`pathlib.Path`, :py:class:`dict`)
        """
        if isinstance(b, bytes):

            s = str.rstrip(bytes.decode(b, encoding="utf-8"))

        else:

            s = b

        line = str.rstrip(s)
        facts_found, _, name = str.partition(line, " ")
        entry = {}
        for fact in str.split(facts_found[:-1], ";"):

            key, _, value = str.partition(fact, "=")
            entry[key.lower()] = value

        return pathlib.Path(name), entry


class Client(BaseClient):
    """
    FTP client.

    :param loop: loop to use for creating connection and binding with streams
    :type loop: :py:class:`asyncio.BaseEventLoop`

    :param create_connection: factory for creating
        connection.
        Using default :py:meth:`asyncio.BaseEventLoop.create_connection`
        if omitted.
    :type create_connection: :py:func:`callable`

    :param timeout: timeout for read operations
    :type timeout: :py:class:`float` or :py:class:`int`
    """

    @asyncio.coroutine
    def connect(self, host, port=21):
        """
        :py:func:`asyncio.coroutine`

        Connect to server.

        :param host: host name for connection
        :type host: :py:class:`str`

        :param port: port number for connection
        :type port: :py:class:`int`
        """
        yield from super().connect(host, port)
        code, info = yield from self.command(None, "220", "120")
        return info

    @asyncio.coroutine
    def login(self, user="anonymous", password="anon@", account=""):
        """
        :py:func:`asyncio.coroutine`

        Server authentication.

        :param user: username
        :type user: :py:class:`str`

        :param password: password
        :type password: :py:class:`str`

        :param account: account (almost always blank)
        :type account: :py:class:`str`

        :raises aioftp.StatusCodeError: if unknown code received
        """
        code, info = yield from self.command("USER " + user, ("230", "33x"))
        while code.matches("33x"):

            if code == "331":

                cmd = "PASS " + password

            elif code == "332":

                cmd = "ACCT " + account

            else:

                raise errors.StatusCodeError("33x", code, info)

            code, info = yield from self.command(cmd, ("230", "33x"))

    @asyncio.coroutine
    def get_current_directory(self):
        """
        :py:func:`asyncio.coroutine`

        Getting current working directory.

        :rtype: :py:class:`pathlib.Path`
        """
        code, info = yield from self.command("PWD", "257")
        directory = self.parse_directory_response(info[-1])
        return directory

    @asyncio.coroutine
    def change_directory(self, path=".."):
        """
        :py:func:`asyncio.coroutine`

        Change current directory. Goes «up» if no parameters passed.

        :param pathlib.Path path: new directory, goes «up» if omitted
        :type path: :py:class:`str` or :py:class:`pathlib.Path`
        """
        if path in ("..", pathlib.Path("..")):

            cmd = "CDUP"

        else:

            cmd = "CWD " + str(path)

        yield from self.command(cmd, "2xx")

    @asyncio.coroutine
    def make_directory(self, path, *, parents=True):
        """
        :py:func:`asyncio.coroutine`

        Make directory.

        :param path: path to directory to create
        :type path: :py:class:`str` or :py:class:`pathlib.Path`

        :param parents: create parents if does not exists
        :type parents: :py:class:`bool`
        """
        path = pathlib.Path(path)
        need_create = []
        while path.name and not (yield from self.exists(path)):

            need_create.append(path)
            path = path.parent
            if not parents:

                break

        need_create.reverse()
        for path in need_create:

            yield from self.command("MKD " + str(path), "257")

    @asyncio.coroutine
    def remove_directory(self, path):
        """
        :py:func:`asyncio.coroutine`

        Low level remove method for removing empty directory.

        :param path: empty directory to remove
        :type path: :py:class:`str` or :py:class:`pathlib.Path`
        """
        yield from self.command("RMD " + str(path), "250")

    @asyncio.coroutine
    def list(self, path="", *, recursive=False):
        """
        :py:func:`asyncio.coroutine`

        List all files and directories in "path".

        :param path: directory or file path
        :type path: :py:class:`str` or :py:class:`pathlib.Path`

        :param recursive: list recursively
        :type recursive: :py:class:`bool`

        :rtype: :py:class:`list` or :py:class:`None`
        """
        result = []
        directories = []
        command = str.strip("MLSD " + str(path))
        stream = yield from self.get_stream(command, "1xx")
        while True:

            line = yield from stream.readline()
            if not line:

                yield from stream.finish()
                break

            name, info = self.parse_mlsx_line(line)
            if info["type"] in ("file", "dir"):

                stat = path / name, info
                if info["type"] == "dir":

                    directories.append(stat)

                result.append(stat)

        if recursive:

            for name, info in directories:

                result += yield from self.list(name, recursive=recursive)

        return result

    @asyncio.coroutine
    def stat(self, path):
        """
        :py:func:`asyncio.coroutine`

        Getting path stats.

        :param path: path for getting info
        :type path: :py:class:`str` or :py:class:`pathlib.Path`

        :return: path info
        :rtype: :py:class:`dict`
        """
        code, info = yield from self.command("MLST " + str(path), "2xx")
        name, info = self.parse_mlsx_line(str.lstrip(info[1]))
        return info

    @asyncio.coroutine
    def is_file(self, path):
        """
        :py:func:`asyncio.coroutine`

        Checks if path is file.

        :rtype: :py:class:`bool`
        """
        info = yield from self.stat(path)
        return info["type"] == "file"

    @asyncio.coroutine
    def is_dir(self, path):
        """
        :py:func:`asyncio.coroutine`

        Checks if path is dir.

        :rtype: :py:class:`bool`
        """
        info = yield from self.stat(path)
        return info["type"] == "dir"

    @asyncio.coroutine
    def exists(self, path):
        """
        :py:func:`asyncio.coroutine`

        Check path for existence.

        :param path: path for checking on server side
        :type path: :py:class:`str` or :py:class:`pathlib.Path`

        :rtype: :py:class:`bool`
        """
        code, info = yield from self.command(
            "MLST " + str(path),
            ("2xx", "550")
        )
        exists = code.matches("2xx")
        return exists

    @asyncio.coroutine
    def rename(self, source, destination):
        """
        :py:func:`asyncio.coroutine`

        Rename (move) file or directory.

        :param source: path to rename
        :type source: :py:class:`str` or :py:class:`pathlib.Path`

        :param destination: path new name
        :type destination: :py:class:`str` or :py:class:`pathlib.Path`
        """
        yield from self.command("RNFR " + str(source), "350")
        yield from self.command("RNTO " + str(destination), "2xx")

    @asyncio.coroutine
    def remove_file(self, path):
        """
        :py:func:`asyncio.coroutine`

        Low level remove method for removing file.

        :param path: file to remove
        :type path: :py:class:`str` or :py:class:`pathlib.Path`
        """
        yield from self.command("DELE " + str(path), "2xx")

    @asyncio.coroutine
    def remove(self, path):
        """
        :py:func:`asyncio.coroutine`

        High level remove method for removing path recursively (file or
        directory).

        :param path: path to remove
        :type path: :py:class:`str` or :py:class:`pathlib.Path`
        """
        if (yield from self.exists(path)):

            info = yield from self.stat(path)
            if info["type"] == "file":

                yield from self.remove_file(path)

            elif info["type"] == "dir":

                for name, info in (yield from self.list(path)):

                    if info["type"] in ("dir", "file"):

                        yield from self.remove(name)

                yield from self.remove_directory(path)

    @asyncio.coroutine
    def upload_stream(self, destination):
        """
        :py:func:`asyncio.coroutine`

        Create stream for write data to `destination` file.

        :param destination: destination path of file on server side
        :type destination: :py:class:`str` or :py:class:`pathlib.Path`

        :rtype: :py:class:`aioftp.StreamIO`
        """
        return (yield from self.get_stream("STOR " + str(destination), "1xx"))

    @asyncio.coroutine
    def append_stream(self, destination):
        """
        :py:func:`asyncio.coroutine`

        Create stream for append (write) data to `destination` file.

        :param destination: destination path of file on server side
        :type destination: :py:class:`str` or :py:class:`pathlib.Path`

        :rtype: :py:class:`aioftp.StreamIO`
        """
        return (yield from self.get_stream("APPE " + str(destination), "1xx"))

    @asyncio.coroutine
    def upload(self, source, destination="", *, write_into=False,
               block_size=common.default_block_size):
        """
        :py:func:`asyncio.coroutine`

        High level upload method for uploading files and directories
        recursively from file system.

        :param source: source path of file or directory on client side
        :type source: :py:class:`str` or :py:class:`pathlib.Path`

        :param destination: destination path of file or directory on server
            side
        :type destination: :py:class:`str` or :py:class:`pathlib.Path`

        :param write_into: write source into destination (if you want upload
            file and change it name, as well with directories)
        :type write_into: :py:class:`bool`

        :param block_size: block size for transaction
        :type block_size: :py:class:`int`
        """
        source = pathlib.Path(source)
        destination = pathlib.Path(destination)
        if not write_into:

            destination = destination / source.name

        if source.is_file():

            yield from self.make_directory(destination.parent)
            with source.open(mode="rb") as fin:

                stream = yield from self.upload_stream(destination)
                while True:

                    block = fin.read(block_size)
                    if not block:

                        yield from stream.finish()
                        break

                    yield from stream.write(block)

        elif source.is_dir():

            yield from self.make_directory(destination)
            for p in source.rglob("*"):

                if write_into:

                    relative = destination.name / p.relative_to(source)

                else:

                    relative = p.relative_to(source.parent)

                if p.is_dir():

                    yield from self.make_directory(relative)

                else:

                    yield from self.make_directory(relative.parent)
                    with p.open(mode="rb") as fin:

                        stream = yield from self.upload_stream(relative)
                        while True:

                            block = fin.read(block_size)
                            if not block:

                                yield from stream.finish()
                                break

    @asyncio.coroutine
    def download_stream(self, source):
        """
        :py:func:`asyncio.coroutine`

        Create stream for read data from `source` file.

        :param source: source path of file on server side
        :type source: :py:class:`str` or :py:class:`pathlib.Path`

        :rtype: :py:class:`aioftp.StreamIO`
        """
        return (yield from self.get_stream("RETR " + str(source), "1xx"))

    @asyncio.coroutine
    def download(self, source, destination="", *, write_into=False,
                 block_size=common.default_block_size):
        """
        :py:func:`asyncio.coroutine`

        High level download method for downloading files and directories
        recursively and save them to the file system.

        :param source: source path of file or directory on server side
        :type source: :py:class:`str` or :py:class:`pathlib.Path`

        :param destination: destination path of file or directory on client
            side
        :type destination: :py:class:`str` or :py:class:`pathlib.Path`

        :param write_into: write source into destination (if you want download
            file and change it name, as well with directories)
        :type write_into: :py:class:`bool`

        :param block_size: block size for transaction
        :type block_size: :py:class:`int`
        """
        source = pathlib.Path(source)
        destination = pathlib.Path(destination)
        if not write_into:

            destination = destination / source.name

        if (yield from self.is_file(source)):

            if not destination.parent.exists():

                destination.parent.mkdir()

            with destination.open(mode="wb") as fout:

                stream = yield from self.download_stream(source)
                while True:

                    block = yield from stream.read(block_size)
                    if not block:

                        yield from stream.finish()
                        break

                    fout.write(block)

        elif (yield from self.is_dir(source)):

            if not destination.exists():

                destination.mkdir(parents=True)

            for name, info in (yield from self.list(source, recursive=True)):

                full = destination / name.relative_to(source)
                if info["type"] == "file":

                    with full.open(mode="wb") as fout:

                        stream = yield from self.download_stream(name)
                        while True:

                            block = yield from stream.read(block_size)
                            if not block:

                                yield from stream.finish()
                                break

                            fout.write(block)

                elif info["type"] == "dir":

                    if not full.exists():

                        full.mkdir(parents=True)

    @asyncio.coroutine
    def quit(self):
        """
        :py:func:`asyncio.coroutine`

        Send "QUIT" and close connection.
        """
        yield from self.command("QUIT", "2xx")
        self.close()

    @asyncio.coroutine
    def get_passive_connection(self, conn_type="I"):
        """
        :py:func:`asyncio.coroutine`

        Getting pair of reader, writer for passive connection with server.

        :param conn_type: connection type ("I", "A", "E", "L")
        :type conn_type: :py:class:`str`

        :rtype: (:py:class:`asyncio.StreamReader`,
            :py:class:`asyncio.StreamWriter`)
        """
        yield from self.command("TYPE " + conn_type, "200")
        code, info = yield from self.command("PASV", "227")
        ip, port = self.parse_address_response(info[-1])
        reader, writer = yield from open_connection(
            ip,
            port,
            self.loop,
            self.create_connection,
            read_speed_limit=self.read_speed_limit,
            read_memory=self.read_memory,
            write_speed_limit=self.write_speed_limit,
            write_memory=self.write_memory
        )
        return reader, writer

    @asyncio.coroutine
    def get_stream(self, *command_args, conn_type="I"):
        """
        :py:func:`asyncio.coroutine`

        Create :py:class:`aioftp.StreamIO` for straight read/write io.

        :param command_args: arguments for :py:meth:`aioftp.Client.command`

        :param conn_type: connection type ("I", "A", "E", "L")
        :type conn_type: :py:class:`str`

        :rtype: :py:class:`aioftp.StreamIO`
        """
        reader, writer = yield from self.get_passive_connection(conn_type)
        yield from self.command(*command_args)
        stream = StreamIO(self, reader, writer)
        return stream

    @asyncio.coroutine
    def abort(self):
        """
        :py:func:`asyncio.coroutine`

        Request data transfer abort.
        """
        yield from self.command("ABOR")



from common import *


@nose.tools.raises(OSError)
@aioftp_setup()
def test_client_without_server(loop, client, server):

    yield from client.connect("127.0.0.1", PORT)


@aioftp_setup()
@with_connection
def test_connection(loop, client, server):

    pass


@aioftp_setup()
@with_connection
def test_quit(loop, client, server):

    yield from client.quit()


@aioftp_setup()
@with_connection
def test_dirty_quit(loop, client, server):

    yield from asyncio.sleep(0.5, loop=loop)
    client.close()
    yield from asyncio.sleep(0.5, loop=loop)


@aioftp_setup()
@with_connection
@expect_codes_in_exception("502")
def test_not_implemented(loop, client, server):

    yield from client.command("FOOBAR", "2xx", "1xx")


@aioftp_setup()
@with_connection
@expect_codes_in_exception("502")
def test_type_not_implemented(loop, client, server):

    yield from client.login()
    yield from client.get_passive_connection("A")


@nose.tools.timed(2)
@nose.tools.raises(ConnectionResetError)
@aioftp_setup()
@with_connection
def test_extra_pasv_connection(loop, client, server):

    yield from client.login()
    r, w = yield from client.get_passive_connection()
    er, ew = yield from client.get_passive_connection()
    while True:

        ew.write(b"-" * 8192)
        yield from asyncio.sleep(0.1, loop=loop)
        yield from ew.drain()

    yield from client.quit()


@nose.tools.timed(2)
@nose.tools.raises(ConnectionRefusedError)
@aioftp_setup()
@with_connection
def test_closing_pasv_connection(loop, client, server):

    yield from client.login()
    r, w = yield from client.get_passive_connection()
    host, port = w.transport.get_extra_info("peername")
    yield from client.quit()
    yield from asyncio.open_connection(host, port, loop=loop)


@nose.tools.timed(2)
@nose.tools.raises(ConnectionResetError)
@aioftp_setup()
@with_connection
def test_server_shutdown(loop, client, server):

    @asyncio.coroutine
    def close_server():

        yield from asyncio.sleep(1, loop=loop)
        server.close()
        yield from server.wait_closed()

    yield from client.login()
    asyncio.async(close_server(), loop=loop)
    while True:

        yield from client.list()
        yield from asyncio.sleep(0.5, loop=loop)


if __name__ == "__main__":

    import logging
    import os

    os.environ["AIOFTP_TESTS"] = "PathIO"
    os.environ["PYTHONASYNCIODEBUG"] = "1"
    logging.basicConfig(
        level=logging.INFO
    )

    test_server_shutdown()
    print("done")



import pathlib

import nose

from common import *


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo"),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_exists(loop, client, server, *, tmp_dir):

    yield from client.login()
    yield from client.make_directory("bar")
    exists = yield from client.exists("bar")
    nose.tools.eq_(exists, True)

    yield from client.remove_directory("bar")
    exists = yield from client.exists("bar")
    nose.tools.eq_(exists, False)


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo"),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_is_dir(loop, client, server, *, tmp_dir):

    yield from client.login()
    yield from client.make_directory("bar")
    is_dir = yield from client.is_dir("bar")
    nose.tools.eq_(is_dir, True)

    tmp_file = tmp_dir / "foo.txt"
    tmp_file.touch()

    is_dir = yield from client.is_dir("foo.txt")
    nose.tools.eq_(is_dir, False)

    yield from client.remove_directory("bar")

    tmp_file.unlink()


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo"),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_is_file(loop, client, server, *, tmp_dir):

    yield from client.login()
    yield from client.make_directory("bar")
    is_file = yield from client.is_file("bar")
    nose.tools.eq_(is_file, False)

    tmp_file = tmp_dir / "foo.txt"
    tmp_file.touch()

    is_file = yield from client.is_file("foo.txt")
    nose.tools.eq_(is_file, True)

    yield from client.remove_directory("bar")

    tmp_file.unlink()


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo"),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_stats(loop, client, server, *, tmp_dir):

    s = "i'm a text"
    tmp_file = tmp_dir / "foo.txt"
    with tmp_file.open(mode="w") as fout:

        fout.write(s)

    yield from client.login()
    stats = yield from client.stat("foo.txt")
    nose.tools.eq_(stats["type"], "file")
    nose.tools.eq_(stats["size"], str(len(s)))

    tmp_file.unlink()


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo"),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_native_list(loop, client, server, *, tmp_dir):

    tmp_file = tmp_dir / "foo.txt"
    tmp_file.touch()

    lines = []
    yield from client.login()
    stream = yield from client.get_stream(str.strip("LIST"), "1xx")
    while True:

        raw = yield from stream.readline()
        if not raw:

            break

        lines.append(raw)

    con = next(iter(server.connections.values()))
    yield from client.quit()
    ans = yield from server.build_list_string(con, tmp_file)
    tmp_file.unlink()

    nose.tools.eq_(len(lines), 1)
    line, *_ = lines
    line = str.strip(bytes.decode(line, "utf-8"))
    nose.tools.eq_(line, ans)


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo"),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_recursive_list(loop, client, server, *, tmp_dir):

    d = tmp_dir / "foo"
    (tmp_dir / "bar.txt").touch()
    d.mkdir()
    (d / "baz.foo").touch()

    yield from client.login()
    paths = set()
    for path, stats in (yield from client.list(recursive=True)):

        paths.add(path)

    (d / "baz.foo").unlink()
    (tmp_dir / "bar.txt").unlink()
    d.rmdir()
    names = "bar.txt", "foo", "foo/baz.foo"
    nose.tools.ok_(paths == set(map(pathlib.Path, names)))

    yield from client.quit()


@nose.tools.raises(aioftp.PathIsNotFileOrDir)
@aioftp_setup()
def test_not_a_file_or_dir(loop, client, server):

    class NotFileOrDirPathIO(aioftp.AbstractPathIO):

        @asyncio.coroutine
        def is_file(self, path):

            return False

        @asyncio.coroutine
        def is_dir(self, path):

            return False

    yield from server.build_mlsx_string(
        aioftp.Connection(
            path_io=NotFileOrDirPathIO(loop),
            path_timeout=None,
            loop=loop,
        ),
        pathlib.Path("/foo")
    )


if __name__ == "__main__":

    import logging
    import os

    os.environ["AIOFTP_TESTS"] = "PathIO"
    logging.basicConfig(
        level=logging.INFO
    )

    test_exists()
    test_is_dir()
    test_is_file()
    test_stats()
    test_native_list()
    test_recursive_list()
    test_not_a_file_or_dir()
    print("done")



import io

import nose

from common import *


@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="/"),)],
        {"path_io_factory": aioftp.MemoryPathIO}))
@with_connection
def test_memory_path_upload_download(loop, client, server):

    yield from client.login()
    r = yield from client.list()
    nose.tools.eq_(len(r), 0)

    data = b"foobar"

    stream = yield from client.upload_stream("foo.txt")
    yield from stream.write(data)
    yield from stream.finish()

    stream = yield from client.download_stream("foo.txt")
    rdata = yield from stream.read()
    yield from stream.finish()


    yield from client.quit()
    nose.tools.eq_(data, rdata)


@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="/"),)],
        {"path_io_factory": aioftp.MemoryPathIO}))
@with_connection
def test_memory_path_directory_create(loop, client, server):

    yield from client.login()
    r = yield from client.list()
    nose.tools.eq_(len(r), 0)

    yield from client.make_directory("bar")
    data = b"foobar"

    stream = yield from client.upload_stream("bar/foo.txt")
    yield from stream.write(data)
    yield from stream.finish()

    stream = yield from client.download_stream("bar/foo.txt")
    rdata = yield from stream.read()
    yield from stream.finish()

    nose.tools.eq_(data, rdata)


@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="/"),)],
        {"path_io_factory": aioftp.MemoryPathIO}))
@with_connection
def test_memory_path_append(loop, client, server):

    yield from client.login()
    r = yield from client.list()
    nose.tools.eq_(len(r), 0)

    data = b"foobar"

    stream = yield from client.upload_stream("foo.txt")
    yield from stream.write(data)
    yield from stream.finish()

    stream = yield from client.append_stream("foo.txt")
    yield from stream.write(data)
    yield from stream.finish()

    stream = yield from client.download_stream("foo.txt")
    rdata = yield from stream.read()
    yield from stream.finish()

    yield from client.quit()
    nose.tools.eq_(data * 2, rdata)


@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="/"),)],
        {"path_io_factory": aioftp.MemoryPathIO}))
@with_connection
def test_memory_path_remove(loop, client, server):

    yield from client.login()
    r = yield from client.list()
    nose.tools.eq_(len(r), 0)

    yield from client.make_directory("bar")
    data = b"foobar"

    stream = yield from client.upload_stream("bar/foo.txt")
    yield from stream.write(data)
    yield from stream.finish()

    yield from client.remove("bar")
    r = yield from client.list()
    nose.tools.eq_(len(r), 0)

    yield from client.quit()


@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="/"),)],
        {"path_io_factory": aioftp.MemoryPathIO}))
@with_connection
@expect_codes_in_exception("550")
def test_memory_path_unreachable_path_upload(loop, client, server):

    yield from client.login()
    data = b"foobar"

    stream = yield from client.upload_stream("bar")
    yield from stream.write(data)
    yield from stream.finish()

    stream = yield from client.upload_stream("bar/foo.txt")
    yield from stream.write(data)
    yield from stream.finish()


@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="/"),)],
        {"path_io_factory": aioftp.MemoryPathIO}))
def test_memory_path_unreachable_path_raw(loop, client, server):

    mp = aioftp.MemoryPathIO(loop)
    nose.tools.eq_(mp.get_node(pathlib.Path("/foo/bar/baz")), None)
    yield from mp.open(pathlib.Path("/foo"), "wb")
    nose.tools.eq_(mp.get_node(pathlib.Path("/foo/bar/baz")), None)


@nose.tools.raises(FileExistsError)
@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="/"),)],
        {"path_io_factory": aioftp.MemoryPathIO}))
def test_memory_path_mkdir_on_exist(loop, client, server):

    mp = aioftp.MemoryPathIO(loop)
    yield from mp.mkdir(pathlib.Path("/foo"))
    yield from mp.mkdir(pathlib.Path("/foo"))


@nose.tools.raises(FileNotFoundError)
@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="/"),)],
        {"path_io_factory": aioftp.MemoryPathIO}))
def test_memory_path_mkdir_unreachable_parents(loop, client, server):

    mp = aioftp.MemoryPathIO(loop)
    yield from mp.mkdir(pathlib.Path("/foo/bar"))


@nose.tools.raises(FileExistsError)
@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="/"),)],
        {"path_io_factory": aioftp.MemoryPathIO}))
def test_memory_path_unreachable_mkdir_cause_file(loop, client, server):

    mp = aioftp.MemoryPathIO(loop)
    yield from mp.open(pathlib.Path("/foo"), "wb")
    yield from mp.mkdir(pathlib.Path("/foo/bar"))


@nose.tools.raises(FileExistsError)
@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="/"),)],
        {"path_io_factory": aioftp.MemoryPathIO}))
def test_memory_path_unreachable_mkdir_cause_file_with_parents(loop, client,
                                                               server):

    mp = aioftp.MemoryPathIO(loop)
    yield from mp.open(pathlib.Path("/foo"), "wb")
    yield from mp.mkdir(pathlib.Path("/foo/bar"), parents=True)


@nose.tools.raises(FileNotFoundError)
@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="/"),)],
        {"path_io_factory": aioftp.MemoryPathIO}))
def test_memory_path_remove_directory_not_exists(loop, client, server):

    mp = aioftp.MemoryPathIO(loop)
    yield from mp.rmdir(pathlib.Path("/foo"))


@nose.tools.raises(NotADirectoryError)
@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="/"),)],
        {"path_io_factory": aioftp.MemoryPathIO}))
def test_memory_path_remove_directory_is_file(loop, client, server):

    mp = aioftp.MemoryPathIO(loop)
    yield from mp.open(pathlib.Path("/foo"), "wb")
    yield from mp.rmdir(pathlib.Path("/foo"))


@nose.tools.raises(OSError)
@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="/"),)],
        {"path_io_factory": aioftp.MemoryPathIO}))
def test_memory_path_remove_directory_not_empty(loop, client, server):

    mp = aioftp.MemoryPathIO(loop)
    yield from mp.mkdir(pathlib.Path("/foo"))
    yield from mp.open(pathlib.Path("/foo/bar"), "wb")
    yield from mp.rmdir(pathlib.Path("/foo"))


@nose.tools.raises(FileNotFoundError)
@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="/"),)],
        {"path_io_factory": aioftp.MemoryPathIO}))
def test_memory_path_unlink_not_exists(loop, client, server):

    mp = aioftp.MemoryPathIO(loop)
    yield from mp.unlink(pathlib.Path("/foo"))


@nose.tools.raises(IsADirectoryError)
@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="/"),)],
        {"path_io_factory": aioftp.MemoryPathIO}))
def test_memory_path_unlink_not_file(loop, client, server):

    mp = aioftp.MemoryPathIO(loop)
    yield from mp.mkdir(pathlib.Path("/foo"))
    yield from mp.unlink(pathlib.Path("/foo"))


@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="/"),)],
        {"path_io_factory": aioftp.MemoryPathIO}))
def test_memory_path_list_not_exists(loop, client, server):

    mp = aioftp.MemoryPathIO(loop)
    r = yield from mp.list(pathlib.Path("/foo"))
    nose.tools.eq_(r, tuple())


@nose.tools.raises(FileNotFoundError)
@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="/"),)],
        {"path_io_factory": aioftp.MemoryPathIO}))
def test_memory_path_stat_not_exists(loop, client, server):

    mp = aioftp.MemoryPathIO(loop)
    r = yield from mp.stat(pathlib.Path("/foo"))


@nose.tools.raises(FileNotFoundError)
@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="/"),)],
        {"path_io_factory": aioftp.MemoryPathIO}))
def test_memory_path_open_read_not_exists(loop, client, server):

    mp = aioftp.MemoryPathIO(loop)
    yield from mp.open(pathlib.Path("/foo"))


@nose.tools.raises(FileNotFoundError)
@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="/"),)],
        {"path_io_factory": aioftp.MemoryPathIO}))
def test_memory_path_open_write_parent_file(loop, client, server):

    mp = aioftp.MemoryPathIO(loop)
    yield from mp.open(pathlib.Path("/foo"), "wb")
    yield from mp.open(pathlib.Path("/foo/bar"), "wb")


@nose.tools.raises(IsADirectoryError)
@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="/"),)],
        {"path_io_factory": aioftp.MemoryPathIO}))
def test_memory_path_open_write_dir(loop, client, server):

    mp = aioftp.MemoryPathIO(loop)
    yield from mp.mkdir(pathlib.Path("/foo"))
    yield from mp.open(pathlib.Path("/foo"), "wb")


@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="/"),)],
        {"path_io_factory": aioftp.MemoryPathIO}))
def test_memory_path_open_write_over(loop, client, server):

    mp = aioftp.MemoryPathIO(loop)
    yield from mp.open(pathlib.Path("/foo"), "wb")
    yield from mp.open(pathlib.Path("/foo"), "wb")


@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="/"),)],
        {"path_io_factory": aioftp.MemoryPathIO}))
def test_memory_path_open_append_new(loop, client, server):

    mp = aioftp.MemoryPathIO(loop)
    yield from mp.open(pathlib.Path("/foo"), "ab")


@nose.tools.raises(IsADirectoryError)
@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="/"),)],
        {"path_io_factory": aioftp.MemoryPathIO}))
def test_memory_path_open_append_over_dir(loop, client, server):

    mp = aioftp.MemoryPathIO(loop)
    yield from mp.mkdir(pathlib.Path("/foo"))
    yield from mp.open(pathlib.Path("/foo"), "ab")


@nose.tools.raises(ValueError)
@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="/"),)],
        {"path_io_factory": aioftp.MemoryPathIO}))
def test_memory_path_open_bad_mode(loop, client, server):

    mp = aioftp.MemoryPathIO(loop)
    yield from mp.open(pathlib.Path("/foo"), "foobar")


@nose.tools.raises(FileNotFoundError)
@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="/"),)],
        {"path_io_factory": aioftp.MemoryPathIO}))
def test_memory_path_rename_not_exists(loop, client, server):

    mp = aioftp.MemoryPathIO(loop)
    yield from mp.rename(pathlib.Path("/foo"), pathlib.Path("/bar"))


@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="/"),)],
        {"path_io_factory": aioftp.MemoryPathIO}))
def test_memory_path_rename_to_not_exists(loop, client, server):

    mp = aioftp.MemoryPathIO(loop)
    s, d = pathlib.Path("/foo"), pathlib.Path("/bar")
    yield from mp.open(s, "wb")
    yield from mp.rename(s, d)
    nose.tools.eq_((yield from mp.exists(d)), True)
    nose.tools.eq_((yield from mp.exists(s)), False)


@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="/"),)],
        {"path_io_factory": aioftp.MemoryPathIO}))
def test_memory_path_rename_to_exists(loop, client, server):

    mp = aioftp.MemoryPathIO(loop)
    s, d = pathlib.Path("/foo"), pathlib.Path("/bar")
    yield from mp.open(s, "wb")
    yield from mp.open(d, "wb")
    yield from mp.rename(s, d)
    nose.tools.eq_((yield from mp.exists(d)), True)
    nose.tools.eq_((yield from mp.exists(s)), False)


if __name__ == "__main__":

    import logging
    import os

    os.environ["AIOFTP_TESTS"] = "PathIO"
    logging.basicConfig(
        level=logging.INFO
    )

    test_memory_path_rename_to_not_exists()
    print("done")



import time

import nose

import aioftp
from common import *


@aioftp_setup(
    client_args=(
        [],
        {
            "read_speed_limit": 100 * 1024  # 100 Kib
        },
    ))
@with_connection
@with_tmp_dir("foo")
def test_client_read_throttle(loop, client, server, *, tmp_dir):

    big_file = tmp_dir / "foo.txt"
    with big_file.open("wb") as fout:

        fout.write(b"-" * (3 * 100 * 1024))  # 300 Kib

    start = time.perf_counter()
    yield from client.login()
    stream = yield from client.download_stream("tests/foo/foo.txt")
    count = 0
    while True:

        data = yield from stream.read()
        if not data:

            yield from stream.finish()
            break

        count += len(data)

    nose.tools.eq_(count, 3 * 100 * 1024)
    nose.tools.ok_(2.5 < (time.perf_counter() - start) < 3.5)
    big_file.unlink()


@aioftp_setup(
    client_args=(
        [],
        {
            "read_speed_limit": 100 * 1024  # 100 Kib
        },
    ))
@with_connection
@with_tmp_dir("foo")
def test_client_write_with_read_throttle(loop, client, server, *, tmp_dir):

    start = time.perf_counter()
    big_file = tmp_dir / "foo.txt"
    yield from client.login()
    stream = yield from client.upload_stream("tests/foo/foo.txt")
    count = 0
    for _ in range(3 * 100):  # 300 Kib

        yield from stream.write(b"-" * 1024)

    yield from stream.finish()

    with big_file.open() as fin:

        data = fin.read()

    nose.tools.eq_(len(data), 3 * 100 * 1024)
    nose.tools.ok_((time.perf_counter() - start) < 0.1)
    big_file.unlink()


@aioftp_setup(
    client_args=(
        [],
        {
            "write_speed_limit": 100 * 1024  # 100 Kib
        },
    ))
@with_connection
@with_tmp_dir("foo")
def test_client_read_with_write_throttle(loop, client, server, *, tmp_dir):

    big_file = tmp_dir / "foo.txt"
    with big_file.open("wb") as fout:

        fout.write(b"-" * (3 * 100 * 1024))  # 300 Kib

    start = time.perf_counter()
    yield from client.login()
    stream = yield from client.download_stream("tests/foo/foo.txt")
    count = 0
    while True:

        data = yield from stream.read()
        if not data:

            yield from stream.finish()
            break

        count += len(data)

    nose.tools.eq_(count, 3 * 100 * 1024)
    nose.tools.ok_((time.perf_counter() - start) < 0.1)
    big_file.unlink()


@aioftp_setup(
    client_args=(
        [],
        {
            "write_speed_limit": 100 * 1024  # 100 Kib
        },
    ))
@with_connection
@with_tmp_dir("foo")
def test_client_write_throttle(loop, client, server, *, tmp_dir):

    start = time.perf_counter()
    big_file = tmp_dir / "foo.txt"
    yield from client.login()
    stream = yield from client.upload_stream("tests/foo/foo.txt")
    count = 0
    for _ in range(3 * 100):  # 300 Kib

        yield from stream.write(b"-" * 1024)

    yield from stream.finish()

    with big_file.open() as fin:

        data = fin.read()

    nose.tools.eq_(len(data), 3 * 100 * 1024)
    nose.tools.ok_(2.5 < (time.perf_counter() - start) < 3.5)
    big_file.unlink()


if __name__ == "__main__":

    import logging
    import os

    os.environ["AIOFTP_TESTS"] = "PathIO"
    logging.basicConfig(
        level=logging.INFO
    )

    test_client_write_throttle()
    print("done")



from common import *


@aioftp_setup()
@with_connection
@expect_codes_in_exception("503")
def test_not_logged_in(loop, client, server):

    cwd = yield from client.get_current_directory()


@aioftp_setup()
@with_connection
def test_anonymous_login(loop, client, server):

    yield from client.login()
    yield from client.quit()


@aioftp_setup()
@with_connection
def test_login_with_login_data(loop, client, server):

    yield from client.login("foo", "bar")
    yield from client.quit()


@aioftp_setup(
    server_args=([(aioftp.User("foo"),)], {}))
@with_connection
def test_login_with_login_and_no_password(loop, client, server):

    yield from client.login("foo")
    yield from client.quit()


@aioftp_setup(
    server_args=([(aioftp.User("foo", "bar"),)], {}))
@with_connection
def test_login_with_login_and_password(loop, client, server):

    yield from client.login("foo", "bar")
    yield from client.quit()


@aioftp_setup(
    server_args=([(aioftp.User("foo", "bar"),)], {}))
@with_connection
@expect_codes_in_exception("530")
def test_login_with_login_and_password_no_such_user(loop, client, server):

    yield from client.login("fo", "bar")
    yield from client.quit()


@aioftp_setup(
    server_args=([(aioftp.User("foo", "bar"),)], {}))
@with_connection
@expect_codes_in_exception("530")
def test_login_with_login_and_password_bad_password(loop, client, server):

    yield from client.login("foo", "baz")
    yield from client.quit()


if __name__ == "__main__":

    import logging
    import os

    os.environ["AIOFTP_TESTS"] = "PathIO"
    logging.basicConfig(
        level=logging.INFO
    )

    test_login_with_login_and_no_password()
    print("done")



"""ftp client/server for asyncio"""


from .client import *
from .server import *
from .errors import *
from .common import *
from .pathio import *


__version__ = "0.2.0"
version = tuple(map(int, str.split(__version__, ".")))

__all__ = (
    client.__all__ +
    server.__all__ +
    errors.__all__ +
    common.__all__ +
    pathio.__all__ +
    (version, __version__)
)



import os
import re
from setuptools import setup, find_packages
from configparser import ConfigParser
from setuptools.command.test import test as TestCommand

config = ConfigParser(default_section='metadata', empty_lines_in_values=False)
config.read(os.path.join(os.path.dirname(__file__), 'flit.ini'))


def read(f):
    return open(os.path.join(os.path.dirname(__file__), f)).read().strip()

try:
    version = re.findall(r"""^__version__ = "([^']+)"\r?$""",
                         read(os.path.join('aioftp', '__init__.py')), re.M)[0]
except IndexError:
    raise RuntimeError('Unable to determine version.')


class NoseTestCommand(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pathlib
        import shutil
        import nose

        for path in pathlib.Path("tests").glob("*"):
            if path.is_dir():
                shutil.rmtree(str(path))

        cov = pathlib.Path(".coverage")
        if cov.exists():
            cov.unlink()

        nose.run(argv=['nosetests'])


setup(
    name=config.defaults().get('module'),
    version=version,
    description=('ftp client/server for asyncio'),
    long_description='\n\n'.join(read('readme.rst')),
    classifiers=config.defaults().get('classifiers').splitlines(),
    author=config.defaults().get('author'),
    author_email=config.defaults().get('author-email'),
    url=config.defaults().get('home-page'),
    license='Apache 2',
    packages=find_packages(),
    install_requires=[],
    tests_require=['nose', 'coverage'],
    cmdclass={'test': NoseTestCommand},
    include_package_data=True
)



import asyncio
import pathlib
import functools
import contextlib
import inspect
import datetime
import socket
import collections
import concurrent


from . import common
from . import errors
from . import pathio


__all__ = (
    "Permission",
    "User",
    "Connection",
    "ConnectionConditions",
    "PathConditions",
    "PathPermissions",
    "Server",
)


def add_prefix(message):

    return str.format("aioftp server: {}", message)


class Permission:
    """
    Path permission

    :param path: path
    :type path: :py:class:`str` or :py:class:`pathlib.PurePosixPath`

    :param readable: is readable
    :type readable: :py:class:`bool`

    :param writable: is writable
    :type writable: :py:class:`bool`
    """

    def __init__(self, path="/", *, readable=True, writable=True):

        self.path = pathlib.PurePosixPath(path)
        self.readable = readable
        self.writable = writable

    def is_parent(self, other):

        try:

            other.relative_to(self.path)
            return True

        except ValueError:

            return False

    def __repr__(self):

        return str.format(
            "{}({!r}, readable={!r}, writable={!r})",
            self.__class__.__name__,
            self.path,
            self.readable,
            self.writable,
        )


class User:
    """
    User description.

    :param login: user login
    :type login: :py:class:`str`

    :param password: user password
    :type password: :py:class:`str`

    :param base_path: real user path for file io operations
    :type base_path: :py:class:`str` or :py:class:`pathlib.Path`

    :param home_path: virtual user path for client representation (must be
        absolute)
    :type home_path: :py:class:`str` or :py:class:`pathlib.PurePosixPath`

    :param permissions: list of path permissions
    :type permissions: :py:class:`tuple` or :py:class:`list` of
        :py:class:`aioftp.Permission`

    :param maximum_connections: Maximum connections per user
    :type maximum_connections: :py:class:`int`
    """

    def __init__(self, login=None, password=None, *,
                 base_path=pathlib.Path("."),
                 home_path=pathlib.PurePosixPath("/"), permissions=None,
                 maximum_connections=None):

        self.login = login
        self.password = password
        self.base_path = pathlib.Path(base_path)
        self.home_path = pathlib.PurePosixPath(home_path)
        if not self.home_path.is_absolute():

            raise errors.PathIsNotAbsolute(home_path)

        self.permissions = permissions or [Permission()]
        self.maximum_connections = maximum_connections

    def get_permissions(self, path):
        """
        Return nearest parent permission for `path`.

        :param path: path which permission you want to know
        :type path: :py:class:`str` or :py:class:`pathlib.PurePosixPath`

        :rtype: :py:class:`aioftp.Permission`
        """
        path = pathlib.PurePosixPath(path)
        parents = filter(lambda p: p.is_parent(path), self.permissions)
        perm = min(
            parents,
            key=lambda p: len(path.relative_to(p.path).parts),
            default=Permission(),
        )
        return perm

    def __repr__(self):

        return str.format(
            "{}({!r}, {!r}, base_path={!r}, home_path={!r}, permissions={!r}, "
            "maximum_connections={!r})",
            self.__class__.__name__,
            self.login,
            self.password,
            self.base_path,
            self.home_path,
            self.permissions,
            self.maximum_connections,
        )


class UserManager:
    """
    Abstract user manager.
    """

    @asyncio.coroutine
    def get_user(self, login):
        """
        :py:func:`asyncio.coroutine`

        Get user and response for USER call

        :param login: user's login
        :type login: :py:class:`str`

        :return: (user, logged, code, info)
        :rtype: (:py:class:`aioftp.User`, :py:class:`bool`, :py:class:`str`,
            :py:class:`str`)
        """
        raise NotImplementedError

    @asyncio.coroutine
    def authenticate(self, user, password):
        """
        :py:func:`asyncio.coroutine`

        Check if user can be authenticated with provided password

        :param user: user
        :type user: :py:class:`aioftp.User`

        :rtype: :py:class:`bool`
        """
        raise NotImplementedError

    @asyncio.coroutine
    def notify_logout(self, user):
        """
        :py:func:`asyncio.coroutine`

        Called when user connection is closed if user was initiated

        :param user: user
        :type user: :py:class:`aioftp.User`
        """
        pass


class MemoryUserManager(UserManager):
    """
    A built-in user manager that keeps predefined set of users in memory.
    """

    def __init__(self, users):

        self.users = users or [User()]
        self.available_connections = dict(
            (user, AvailableConnections(user.maximum_connections))
            for user in self.users
        )

    @asyncio.coroutine
    def get_user(self, login):

        user = None
        for u in self.users:

            if u.login is None and user is None:

                user = u

            elif u.login == login:

                user = u
                break

        if user is None:

            raise IndexError("530", "no such username")

        elif self.available_connections[user].locked():

            template = "too much connections for '{}'"
            message = str.format(template, user.login or "anonymous")
            raise IndexError("530", message)

        elif user.login is None:

            logged, code, info = True, "230", "anonymous login"

        elif user.password is None:

            logged, code, info = True, "230", "login without password"

        else:

            logged, code, info = False, "331", "require password"

        self.available_connections[user].acquire()
        return user, logged, code, info

    @asyncio.coroutine
    def authenticate(self, user, password):

        return user.password == password

    @asyncio.coroutine
    def notify_logout(self, user):

        self.available_connections[user].release()


class Connection(collections.defaultdict):
    """
    Connection state container for transparent work with futures for async
    wait

    :param loop: event loop
    :type loop: :py:class:`asyncio.BaseEventLoop`

    Container based on :py:class:`collections.defaultdict`, which holds
    :py:class:`asyncio.Future` as default factory. There is two layers of
    abstraction:

    * Low level based on simple dictionary keys to attributes mapping and
        available at Connection.future.
    * High level based on futures result and dictionary keys to attributes
        mapping and available at Connection.

    To clarify, here is groups of equal expressions
    ::

        >>> connection.future.foo
        >>> connection["foo"]

        >>> connection.foo
        >>> connection["foo"].result()

        >>> del connection.future.foo
        >>> del connection.foo
        >>> del connection["foo"]
    """

    __slots__ = ("future",)

    class Container:

        def __init__(self, storage):

            self.storage = storage

        def __getattr__(self, name):

            return self.storage[name]

        def __delattr__(self, name):

            self.storage.pop(name)

    def __init__(self, *, loop=None, **kwargs):

        super().__init__(functools.partial(asyncio.Future, loop=loop))
        self.future = Connection.Container(self)

        self["loop"].set_result(loop)
        for k, v in kwargs.items():

            self[k].set_result(v)

    def __getattr__(self, name):

        if name in self:

            return self[name].result()

        else:

            raise AttributeError(str.format("'{}' not in storage", name))

    def __setattr__(self, name, value):

        if name in Connection.__slots__:

            super().__setattr__(name, value)

        else:

            if self[name].done():

                self[name] = self.default_factory()

            self[name].set_result(value)

    def __delattr__(self, name):

        if name in self:

            self.pop(name)

        else:

            super().__delattr__(name)


class AvailableConnections:
    """
    Semaphore-like object. Have no blocks, only raises ValueError on bounds
        crossing. If value is `None` have no limits (bounds checks).

    :param value:
    :type value: :py:class:`int` or `None`
    """

    def __init__(self, value=None):

        self.value = self.maximum_value = value

    def locked(self):
        """
        Returns True if semaphore-like can not be acquired.

        :rtype: :py:class:`bool`
        """
        return self.value == 0

    def acquire(self):
        """
        Acquire, decrementing the internal counter by one.
        """
        if self.value is not None:

            self.value -= 1
            if self.value < 0:

                raise ValueError("Too much acquires")

    def release(self):
        """
        Release, incrementing the internal counter by one.
        """
        if self.value is not None:

            self.value += 1
            if self.value > self.maximum_value:

                raise ValueError("Too much releases")


class BaseServer:

    @asyncio.coroutine
    def start(self, host=None, port=0, **kw):
        """
        :py:func:`asyncio.coroutine`

        Start server.

        :param host: ip address to bind for listening.
        :type host: :py:class:`str`

        :param port: port number to bind for listening.
        :type port: :py:class:`int`

        :param kw: keyword arguments, they passed to
            :py:func:`asyncio.start_server`
        """
        self.connections = {}
        self.server_host = host
        self.server_port = port
        self.server = yield from asyncio.start_server(
            self.dispatcher,
            host,
            port,
            loop=self.loop,
            **kw
        )

        for sock in self.server.sockets:

            if sock.family == socket.AF_INET:

                host, port = sock.getsockname()
                message = str.format("serving on {}:{}", host, port)
                common.logger.info(add_prefix(message))

    def close(self):
        """
        Shutdown the server and close all connections. Use this method with
        :py:meth:`aioftp.Server.wait_closed`
        """
        self.server.close()
        for connection in self.connections.values():

            connection._dispatcher.cancel()

    @asyncio.coroutine
    def wait_closed(self):
        """
        :py:func:`asyncio.coroutine`

        Wait server to stop.
        """
        yield from self.server.wait_closed()

    @asyncio.coroutine
    def write_line(self, reader, writer, line, encoding="utf-8", *,
                   loop, socket_timeout=None):

        common.logger.info(add_prefix(line))
        message = line + common.end_of_line
        writer.write(str.encode(message, encoding=encoding))
        yield from asyncio.wait_for(
            writer.drain(),
            socket_timeout,
            loop=loop,
        )

    @asyncio.coroutine
    def write_response(self, reader, writer, code, lines="", list=False, *,
                       loop, socket_timeout=None):
        """
        :py:func:`asyncio.coroutine`

        Complex method for sending response.

        :param reader: connection steram reader
        :type reader: :py:class:`asyncio.StreamReader`

        :param writer: connection stream writer
        :type writer: :py:class:`asyncio.StreamWriter`

        :param code: server response code
        :type code: :py:class:`str`

        :param lines: line or lines, which are response information
        :type lines: :py:class:`str` or :py:class:`collections.Iterable`

        :param list: if true, then lines will be sended without code prefix.
            This is useful for **LIST** FTP command and some others.
        :type list: :py:class:`bool`

        :param loop: event loop
        :type loop: :py:class:`asyncio.BaseEventLoop`

        :param socket_timeout: timeout for socket write operations
        :type socket_timeout: :py:class:`float` or :py:class:`int`
        """
        lines = common.wrap_with_container(lines)
        write = functools.partial(
            self.write_line,
            reader,
            writer,
            loop=loop,
            socket_timeout=socket_timeout,
        )
        if list:

            head, *body, tail = lines
            yield from write(code + "-" + head)
            for line in body:

                yield from write(" " + line)

            yield from write(code + " " + tail)

        else:

            *body, tail = lines
            for line in body:

                yield from write(code + "-" + line)

            yield from write(code + " " + tail)

    @asyncio.coroutine
    def parse_command(self, reader, writer, *, loop, idle_timeout=None):
        """
        :py:func:`asyncio.coroutine`

        Complex method for getting command.

        :param reader: connection steram reader
        :type reader: :py:class:`asyncio.StreamReader`

        :param writer: connection stream writer
        :type writer: :py:class:`asyncio.StreamWriter`

        :param loop: event loop
        :type loop: :py:class:`asyncio.BaseEventLoop`

        :param idle_timeout: timeout for socket read operations, another
            words: how long user can keep silence without sending commands
        :type idle_timeout: :py:class:`float` or :py:class:`int`

        :return: (code, rest)
        :rtype: (:py:class:`str`, :py:class:`str`)
        """
        line = yield from asyncio.wait_for(
            reader.readline(),
            idle_timeout,
            loop=loop,
        )
        if not line:

            raise ConnectionResetError

        s = str.rstrip(bytes.decode(line, encoding="utf-8"))
        common.logger.info(add_prefix(s))
        cmd, _, rest = str.partition(s, " ")
        return str.lower(cmd), rest

    def CommandParser(self, connection):
        """
        Shorthand for parse_command with current connection

        :param connection:
        :type connection: :py:class:`aioftp.Connection`

        :return: :py:meth:`aioftp.Server.parse_command` coroutine
        :rtype: :py:func:`asyncio.coroutine`
        """
        return self.parse_command(
            *connection.command_connection,
            loop=connection.loop,
            idle_timeout=connection.idle_timeout
        )

    @asyncio.coroutine
    def response_writer(self, connection, response_queue):
        """
        :py:func:`asyncio.coroutine`

        Worker for write_response with current connection. Get data to response
        from queue, this is for right order of responses. Exits if received
        `None`.

        :param connection:
        :type connection: :py:class:`aioftp.Connection`

        :param response_queue:
        :type response_queue: :py:class:`asyncio.Queue`

        :return: :py:meth:`aioftp.Server.write_response` coroutine
        :rtype: :py:func:`asyncio.coroutine`
        """
        while True:

            args = yield from response_queue.get()
            if args is None:

                break

            yield from self.write_response(
                *(connection.command_connection + args),
                loop=connection.loop,
                socket_timeout=connection.socket_timeout
            )

    def _generate_tasks(self, connection, pending):

        for o in pending | connection.extra_workers:

            if asyncio.iscoroutine(o):

                yield connection.loop.create_task(o)

            else:

                yield o

        connection.extra_workers.clear()

    @asyncio.coroutine
    def dispatcher(self, reader, writer):

        host, port = writer.transport.get_extra_info("peername", ("", ""))
        message = str.format("new connection from {}:{}", host, port)
        common.logger.info(add_prefix(message))

        key = reader, writer
        response_queue = asyncio.Queue(loop=self.loop)
        connection = Connection(
            client_host=host,
            client_port=port,
            server_host=self.server_host,
            server_port=self.server_port,
            command_connection=(reader, writer),
            socket_timeout=self.socket_timeout,
            path_timeout=self.path_timeout,
            idle_timeout=self.idle_timeout,
            wait_future_timeout=self.wait_future_timeout,
            block_size=self.block_size,
            path_io=self.path_io,
            loop=self.loop,
            extra_workers=set(),
            response=lambda *args: response_queue.put_nowait(args),
            acquired=False,
            _dispatcher=asyncio.Task.current_task(loop=self.loop),
        )

        pending = {
            self.greeting(connection, ""),
            self.response_writer(connection, response_queue),
            self.CommandParser(connection),
        }
        self.connections[key] = connection

        try:

            ok = True
            while ok or not response_queue.empty():

                pending = set(self._generate_tasks(connection, pending))
                done, pending = yield from asyncio.wait(
                    pending,
                    return_when=concurrent.futures.FIRST_COMPLETED,
                    loop=connection.loop
                )
                for task in done:

                    result = task.result()

                    # this is "command" result
                    if isinstance(result, bool):

                        ok = result
                        if not ok:

                            # bad solution, but have no better ideas right now
                            connection.response(None)
                            break

                    # this is parse_command result
                    else:

                        pending.add(self.CommandParser(connection))

                        cmd, rest = result
                        if cmd == "pass":

                            # is there a better solution?
                            cmd = "pass_"

                        if hasattr(self, cmd):

                            pending.add(getattr(self, cmd)(connection, rest))

                        else:

                            message = str.format("'{}' not implemented", cmd)
                            connection.response("502", message)

        finally:

            message = str.format("closing connection from {}:{}", host, port)
            common.logger.info(add_prefix(message))

            if not connection.loop.is_closed():

                for task in pending:

                    if isinstance(task, asyncio.Task):

                        task.cancel()

                if connection.future.passive_server.done():

                    connection.passive_server.close()

                if connection.future.data_connection.done():

                    r, w = connection.data_connection
                    w.close()

                writer.close()

            if connection.acquired:

                self.available_connections.release()

            if connection.future.user.done():

                yield from self.user_manager.notify_logout(connection.user)

            self.connections.pop(key)

    @asyncio.coroutine
    def greeting(self, connection, rest):

        raise NotImplementedError


class ConnectionConditions:
    """
    Decorator for checking `connection` keys for existence or wait for them.
    Available options:

    :param fields: * `ConnectionConditions.user_required` — required "user"
          key, user already identified
        * `ConnectionConditions.login_required` — required "logged" key, user
          already logged in.
        * `ConnectionConditions.passive_server_started` — required
          "passive_server" key, user already send PASV and server awaits
          incomming connection
        * `ConnectionConditions.data_connection_made` — required
          "data_connection" key, user already connected to passive connection
        * `ConnectionConditions.rename_from_required` — required "rename_from"
          key, user already tell filename for rename

    :param wait: Indicates if should wait for parameters for
        `connection.wait_future_timeout`
    :type wait: :py:class:`bool`

    :param fail_code: return code if failure
    :type fail_code: :py:class:`str`

    :param fail_info: return information string if failure. If `None`, then
        use default string
    :type fail_info: :py:class:`str`

    ::

        >>> @ConnectionConditions(
        ...     ConnectionConditions.login_required,
        ...     ConnectionConditions.passive_server_started,
        ...     ConnectionConditions.data_connection_made,
        ...     wait=True)
        ... def foo(self, connection, rest):
        ...     ...
    """
    user_required = ("user", "no user (use USER firstly)")
    login_required = ("logged", "not logged in")
    passive_server_started = (
        "passive_server",
        "no listen socket created (use PASV firstly)"
    )
    data_connection_made = ("data_connection", "no data connection made")
    rename_from_required = ("rename_from", "no filename (use RNFR firstly)")

    def __init__(self, *fields, wait=False, fail_code="503", fail_info=None):

        self.fields = fields
        self.wait = wait
        self.fail_code = fail_code
        self.fail_info = fail_info

    def __call__(self, f):

        @asyncio.coroutine
        @functools.wraps(f)
        def wrapper(cls, connection, rest, *args):

            futures = {connection[name]: msg for name, msg in self.fields}
            aggregate = asyncio.gather(*futures, loop=connection.loop)

            if self.wait:

                timeout = connection.wait_future_timeout

            else:

                timeout = 0

            try:

                yield from asyncio.wait_for(
                    asyncio.shield(aggregate, loop=connection.loop),
                    timeout,
                    loop=connection.loop
                )

            except asyncio.TimeoutError:

                for future, message in futures.items():

                    if not future.done():

                        if self.fail_info is None:

                            template = "bad sequence of commands ({})"
                            info = str.format(template, message)

                        else:

                            info = self.fail_info

                        connection.response(self.fail_code, info)
                        return True

            return (yield from f(cls, connection, rest, *args))

        return wrapper


class PathConditions:
    """
    Decorator for checking paths. Available options:

    * `path_must_exists`
    * `path_must_not_exists`
    * `path_must_be_dir`
    * `path_must_be_file`

    ::

        >>> @PathConditions(
        ...     PathConditions.path_must_exists,
        ...     PathConditions.path_must_be_dir)
        ... def foo(self, connection, path):
        ...     ...
    """
    path_must_exists = ("exists", False, "path does not exists")
    path_must_not_exists = ("exists", True, "path already exists")
    path_must_be_dir = ("is_dir", False, "path is not a directory")
    path_must_be_file = ("is_file", False, "path is not a file")

    def __init__(self, *conditions):

        self.conditions = conditions

    def __call__(self, f):

        @asyncio.coroutine
        @functools.wraps(f)
        def wrapper(cls, connection, rest, *args):

            real_path, virtual_path = cls.get_paths(connection, rest)
            for name, fail, message in self.conditions:

                result = yield from asyncio.wait_for(
                    getattr(connection.path_io, name)(real_path),
                    connection.path_timeout,
                    loop=connection.loop
                )
                if result == fail:

                    connection.response("550", message)
                    return True

            return (yield from f(cls, connection, rest, *args))

        return wrapper


class PathPermissions:
    """
    Decorator for checking path permissions. There is two permissions right
    now:

    * `PathPermissions.readable`
    * `PathPermissions.writable`

    Decorator will check the permissions and return proper code and information
    to client if permission denied

    ::

        >>> @PathPermissions(
        ...     PathPermissions.readable,
        ...     PathPermissions.writable)
        ... def foo(self, connection, path):
        ...     ...
    """
    readable = "readable"
    writable = "writable"

    def __init__(self, *permissions):

        self.permissions = permissions

    def __call__(self, f):

        @asyncio.coroutine
        @functools.wraps(f)
        def wrapper(cls, connection, rest, *args):

            real_path, virtual_path = cls.get_paths(connection, rest)
            current_permission = connection.user.get_permissions(virtual_path)
            for permission in self.permissions:

                if not getattr(current_permission, permission):

                    connection.response("550", "permission denied")
                    return True

                return (yield from f(cls, connection, rest, *args))

        return wrapper


class Server(BaseServer):
    """
    FTP server.

    :param users: list of users or user manager object
    :type users: :py:class:`tuple` or :py:class:`list` of
        :py:class:`aioftp.User` or :py:class:`aioftp.server.UserManager`

    :param loop: loop to use for creating connection and binding with streams
    :type loop: :py:class:`asyncio.BaseEventLoop`

    :param block_size: bytes count for socket read operations
    :type block_size: :py:class:`int`

    :param socket_timeout: timeout for socket read and write operations
    :type socket_timeout: :py:class:`float`, :py:class:`int` or
        :py:class:`None`

    :param path_timeout: timeout for path-related operations (make directory,
        unlink file, etc.)
    :type path_timeout: :py:class:`float`, :py:class:`int` or
        :py:class:`None`

    :param idle_timeout: timeout for socket read operations, another
        words: how long user can keep silence without sending commands
    :type idle_timeout: :py:class:`float`, :py:class:`int` or
        :py:class:`None`

    :param wait_future_timeout: wait for data connection to establish
    :type wait_future_timeout: :py:class:`float`, :py:class:`int` or
        :py:class:`None`

    :param path_io_factory: factory of «path abstract layer»
    :type path_io_factory: :py:class:`aioftp.AbstractPathIO`

    :param maximum_connections: Maximum command connections per server
    :type maximum_connections: :py:class:`int`
    """
    path_facts = (
        ("st_size", "Size"),
        ("st_mtime", "Modify"),
        ("st_ctime", "Create"),
    )

    def __init__(self, users=None, *, loop=None,
                 block_size=common.default_block_size, socket_timeout=None,
                 path_timeout=None, idle_timeout=None, wait_future_timeout=1,
                 path_io_factory=pathio.AsyncPathIO, maximum_connections=None):

        if isinstance(users, UserManager):
            self.user_manager = users
        else:
            self.user_manager = MemoryUserManager(users)

        self.loop = loop or asyncio.get_event_loop()
        self.block_size = block_size
        self.socket_timeout = socket_timeout
        self.path_timeout = path_timeout
        self.idle_timeout = idle_timeout
        self.wait_future_timeout = wait_future_timeout
        self.path_io = path_io_factory(self.loop)

        self.maximum_connections = maximum_connections
        self.available_connections = AvailableConnections(maximum_connections)

    def get_paths(self, connection, path):
        """
        Return *real* and *virtual* paths, resolves ".." with "up" action.
        *Real* path is path for path_io, when *virtual* deals with
        "user-view" and user requests

        :param connection: internal options for current connected user
        :type connection: :py:class:`dict`

        :param path: received path from user
        :type path: :py:class:`str` or :py:class:`pathlib.PurePosixPath`

        :return: (real_path, virtual_path)
        :rtype: (:py:class:`pathlib.Path`, :py:class:`pathlib.PurePosixPath`)
        """
        virtual_path = pathlib.PurePosixPath(path)
        if not virtual_path.is_absolute():

            virtual_path = connection.current_directory / virtual_path

        resolved_virtual_path = pathlib.PurePosixPath("/")
        for part in virtual_path.parts[1:]:

            if part == "..":

                resolved_virtual_path = resolved_virtual_path.parent

            else:

                resolved_virtual_path /= part

        base_path = connection.user.base_path
        real_path = base_path / resolved_virtual_path.relative_to("/")
        return real_path, resolved_virtual_path

    @asyncio.coroutine
    def greeting(self, connection, rest):

        if self.available_connections.locked():

            ok, code, info = False, "421", "Too many connections"

        else:

            ok, code, info = True, "220", "welcome"
            connection.acquired = True
            self.available_connections.acquire()

        connection.response(code, info)
        return ok

    @asyncio.coroutine
    def user(self, connection, rest):

        ok = False
        try:

            args = yield from self.user_manager.get_user(rest)
            connection.user, connection.logged, code, info = args
            connection.current_directory = connection.user.home_path
            ok = True

        except IndexError as e:

            code, info = e.args

        connection.response(code, info)
        return ok

    @ConnectionConditions(ConnectionConditions.user_required)
    @asyncio.coroutine
    def pass_(self, connection, rest):

        if (yield from self.user_manager.authenticate(connection.user, rest)):

            connection.logged = True
            code, info = "230", "normal login"

        else:

            code, info = "530", "wrong password"

        connection.response(code, info)
        return True

    @asyncio.coroutine
    def quit(self, connection, rest):

        connection.response("221", "bye")
        return False

    @ConnectionConditions(ConnectionConditions.login_required)
    @asyncio.coroutine
    def pwd(self, connection, rest):

        code, info = "257", str.format("\"{}\"", connection.current_directory)
        connection.response(code, info)
        return True

    @ConnectionConditions(ConnectionConditions.login_required)
    @PathConditions(
        PathConditions.path_must_exists,
        PathConditions.path_must_be_dir)
    @PathPermissions(PathPermissions.readable)
    @asyncio.coroutine
    def cwd(self, connection, rest):

        real_path, virtual_path = self.get_paths(connection, rest)
        connection.current_directory = virtual_path
        connection.response("250", "")
        return True

    @ConnectionConditions(ConnectionConditions.login_required)
    @asyncio.coroutine
    def cdup(self, connection, rest):

        parent = connection.current_directory.parent
        return (yield from self.cwd(connection, parent))

    @ConnectionConditions(ConnectionConditions.login_required)
    @PathConditions(PathConditions.path_must_not_exists)
    @PathPermissions(PathPermissions.writable)
    @asyncio.coroutine
    def mkd(self, connection, rest):

        real_path, virtual_path = self.get_paths(connection, rest)
        yield from asyncio.wait_for(
            connection.path_io.mkdir(real_path, parents=True),
            connection.path_timeout,
            loop=connection.loop,
        )
        connection.response("257", "")
        return True

    @ConnectionConditions(ConnectionConditions.login_required)
    @PathConditions(
        PathConditions.path_must_exists,
        PathConditions.path_must_be_dir)
    @PathPermissions(PathPermissions.writable)
    @asyncio.coroutine
    def rmd(self, connection, rest):

        real_path, virtual_path = self.get_paths(connection, rest)
        yield from asyncio.wait_for(
            connection.path_io.rmdir(real_path),
            connection.path_timeout,
            loop=connection.loop,
        )
        connection.response("250", "")
        return True

    @asyncio.coroutine
    def build_mlsx_string(self, connection, path):

        stats = {}
        if (yield from asyncio.wait_for(connection.path_io.is_file(path),
                                        connection.path_timeout,
                                        loop=connection.loop)):

            stats["Type"] = "file"

        elif (yield from asyncio.wait_for(connection.path_io.is_dir(path),
                                          connection.path_timeout,
                                          loop=connection.loop)):

            stats["Type"] = "dir"

        else:

            raise errors.PathIsNotFileOrDir(path)

        raw = yield from asyncio.wait_for(
            connection.path_io.stat(path),
            connection.path_timeout,
            loop=connection.loop,
        )
        for attr, fact in Server.path_facts:

            stats[fact] = getattr(raw, attr)

        s = ""
        for fact, value in stats.items():

            s += str.format("{}={};", fact, value)

        s += " " + path.name
        return s

    @ConnectionConditions(
        ConnectionConditions.login_required,
        ConnectionConditions.passive_server_started)
    @PathConditions(PathConditions.path_must_exists)
    @PathPermissions(PathPermissions.readable)
    @asyncio.coroutine
    def mlsd(self, connection, rest):

        @ConnectionConditions(
            ConnectionConditions.data_connection_made,
            wait=True,
            fail_code="425",
            fail_info="Can't open data connection")
        @asyncio.coroutine
        def mlsd_worker(self, connection, rest):

            data_reader, data_writer = connection.data_connection
            del connection.data_connection
            with contextlib.closing(data_writer) as data_writer:

                paths = yield from asyncio.wait_for(
                    connection.path_io.list(real_path),
                    connection.path_timeout,
                    loop=connection.loop,
                )
                for path in paths:

                    s = yield from self.build_mlsx_string(connection, path)
                    message = s + common.end_of_line
                    data_writer.write(str.encode(message, "utf-8"))
                    yield from asyncio.wait_for(
                        data_writer.drain(),
                        connection.socket_timeout,
                        loop=connection.loop,
                    )

            connection.response("200", "mlsd data transfer done")
            return True

        real_path, virtual_path = self.get_paths(connection, rest)
        connection.extra_workers.add(mlsd_worker(self, connection, rest))
        connection.response("150", "mlsd transfer started")
        return True

    @asyncio.coroutine
    def build_list_string(self, connection, path):

        fields = []
        is_dir = yield from asyncio.wait_for(
            connection.path_io.is_dir(path),
            connection.path_timeout,
            loop=connection.loop,
        )
        dir_flag = "d" if is_dir else "-"

        stats = yield from asyncio.wait_for(
            connection.path_io.stat(path),
            connection.path_timeout,
            loop=connection.loop,
        )
        default = list("xwr") * 3
        for i in range(9):

            if (stats.st_mode >> i) & 1 == 0:

                default[i] = "-"

        fields.append(dir_flag + str.join("", reversed(default)))
        fields.append(str(stats.st_nlink))
        fields.append("none")
        fields.append("none")
        fields.append(str(stats.st_size))

        t = datetime.datetime.fromtimestamp(stats.st_ctime)
        fields.append(t.strftime("%b %d %Y"))
        fields.append(path.name)
        s = str.join(" ", fields)
        return s

    @ConnectionConditions(
        ConnectionConditions.login_required,
        ConnectionConditions.passive_server_started)
    @PathConditions(PathConditions.path_must_exists)
    @PathPermissions(PathPermissions.readable)
    @asyncio.coroutine
    def list(self, connection, rest):

        @ConnectionConditions(
            ConnectionConditions.data_connection_made,
            wait=True,
            fail_code="425",
            fail_info="Can't open data connection")
        @asyncio.coroutine
        def list_worker(self, connection, rest):

            data_reader, data_writer = connection.data_connection
            del connection.data_connection
            with contextlib.closing(data_writer) as data_writer:

                paths = yield from asyncio.wait_for(
                    connection.path_io.list(real_path),
                    connection.path_timeout,
                    loop=connection.loop,
                )

                for path in paths:

                    s = yield from self.build_list_string(connection, path)
                    message = s + common.end_of_line
                    data_writer.write(str.encode(message, "utf-8"))
                    yield from asyncio.wait_for(
                        data_writer.drain(),
                        connection.socket_timeout,
                        loop=connection.loop,
                    )

            connection.response("226", "list data transfer done")
            return True

        real_path, virtual_path = self.get_paths(connection, rest)
        connection.extra_workers.add(list_worker(self, connection, rest))
        connection.response("150", "list transfer started")
        return True

    @ConnectionConditions(ConnectionConditions.login_required)
    @PathConditions(PathConditions.path_must_exists)
    @PathPermissions(PathPermissions.readable)
    @asyncio.coroutine
    def mlst(self, connection, rest):

        real_path, virtual_path = self.get_paths(connection, rest)
        s = yield from self.build_mlsx_string(connection, real_path)
        connection.response("250", ["start", s, "end"], True)
        return True

    @ConnectionConditions(ConnectionConditions.login_required)
    @PathConditions(PathConditions.path_must_exists)
    @PathPermissions(PathPermissions.writable)
    @asyncio.coroutine
    def rnfr(self, connection, rest):

        real_path, virtual_path = self.get_paths(connection, rest)
        connection.rename_from = real_path
        connection.response("350", "rename from accepted")
        return True

    @ConnectionConditions(
        ConnectionConditions.login_required,
        ConnectionConditions.rename_from_required)
    @PathConditions(PathConditions.path_must_not_exists)
    @PathPermissions(PathPermissions.writable)
    @asyncio.coroutine
    def rnto(self, connection, rest):

        real_path, virtual_path = self.get_paths(connection, rest)
        rename_from = connection.rename_from
        del connection.rename_from

        yield from asyncio.wait_for(
            connection.path_io.rename(rename_from, real_path),
            connection.path_timeout,
            loop=connection.loop,
        )
        connection.response("250", "")
        return True

    @ConnectionConditions(ConnectionConditions.login_required)
    @PathConditions(
        PathConditions.path_must_exists,
        PathConditions.path_must_be_file)
    @PathPermissions(PathPermissions.writable)
    @asyncio.coroutine
    def dele(self, connection, rest):

        real_path, virtual_path = self.get_paths(connection, rest)
        yield from asyncio.wait_for(
            connection.path_io.unlink(real_path),
            connection.path_timeout,
            loop=connection.loop,
        )
        connection.response("250", "")
        return True

    @ConnectionConditions(
        ConnectionConditions.login_required,
        ConnectionConditions.passive_server_started)
    @PathPermissions(PathPermissions.writable)
    @asyncio.coroutine
    def stor(self, connection, rest, mode="wb"):

        @ConnectionConditions(
            ConnectionConditions.data_connection_made,
            wait=True,
            fail_code="425",
            fail_info="Can't open data connection")
        @asyncio.coroutine
        def stor_worker(self, connection, rest):

            data_reader, data_writer = connection.data_connection
            del connection.data_connection
            try:

                fout = yield from asyncio.wait_for(
                    connection.path_io.open(real_path, mode=mode),
                    connection.path_timeout,
                    loop=connection.loop,
                )
                while True:

                    data = yield from asyncio.wait_for(
                        data_reader.read(connection.block_size),
                        connection.socket_timeout,
                        loop=connection.loop,
                    )
                    if not data:

                        info = "data transfer done"
                        break

                    if connection.future.abort.done() and connection.abort:

                        connection.abort = False
                        info = "data transfer aborted"
                        break

                    yield from asyncio.wait_for(
                        connection.path_io.write(fout, data),
                        connection.path_timeout,
                        loop=connection.loop,
                    )

            finally:

                data_writer.close()
                yield from asyncio.wait_for(
                    connection.path_io.close(fout),
                    connection.path_timeout,
                    loop=connection.loop,
                )

            connection.response("226", info)
            return True

        real_path, virtual_path = self.get_paths(connection, rest)
        parent_is_dir = yield from asyncio.wait_for(
            connection.path_io.is_dir(real_path.parent),
            connection.path_timeout,
            loop=connection.loop,
        )
        if parent_is_dir:

            connection.extra_workers.add(stor_worker(self, connection, rest))
            code, info = "150", "data transfer started"

        else:

            code, info = "550", "path unreachable"

        connection.response(code, info)
        return True

    @ConnectionConditions(
        ConnectionConditions.login_required,
        ConnectionConditions.passive_server_started)
    @PathConditions(
        PathConditions.path_must_exists,
        PathConditions.path_must_be_file)
    @PathPermissions(PathPermissions.readable)
    @asyncio.coroutine
    def retr(self, connection, rest):

        @ConnectionConditions(
            ConnectionConditions.data_connection_made,
            wait=True,
            fail_code="425",
            fail_info="Can't open data connection")
        @asyncio.coroutine
        def retr_worker(self, connection, rest):

            data_reader, data_writer = connection.data_connection
            del connection.data_connection
            try:

                fin = yield from asyncio.wait_for(
                    connection.path_io.open(real_path, mode="rb"),
                    connection.path_timeout,
                    loop=connection.loop,
                )
                while True:

                    data = yield from asyncio.wait_for(
                        connection.path_io.read(fin, connection.block_size),
                        connection.path_timeout,
                        loop=connection.loop,
                    )
                    if not data:

                        info = "data transfer done"
                        break

                    if connection.future.abort.done() and connection.abort:

                        connection.abort = False
                        info = "data transfer aborted"
                        break

                    data_writer.write(data)
                    yield from asyncio.wait_for(
                        data_writer.drain(),
                        connection.socket_timeout,
                        loop=connection.loop,
                    )

            finally:

                data_writer.close()
                yield from asyncio.wait_for(
                    connection.path_io.close(fin),
                    connection.path_timeout,
                    loop=connection.loop,
                )

            connection.response("226", info)
            return True

        real_path, virtual_path = self.get_paths(connection, rest)
        connection.extra_workers.add(retr_worker(self, connection, rest))
        connection.response("150", "data transfer started")
        return True

    @ConnectionConditions(ConnectionConditions.login_required)
    @asyncio.coroutine
    def type(self, connection, rest):

        if rest == "I":

            connection.transfer_type = rest
            code, info = "200", ""

        else:

            code, info = "502", str.format("type '{}' not implemented", rest)

        connection.response(code, info)
        return True

    @ConnectionConditions(ConnectionConditions.login_required)
    @asyncio.coroutine
    def pasv(self, connection, rest):

        @asyncio.coroutine
        def handler(reader, writer):

            if connection.future.data_connection.done():

                writer.close()

            else:

                connection.data_connection = reader, writer

        if not connection.future.passive_server.done():

            connection.passive_server = yield from asyncio.start_server(
                handler,
                connection.server_host,
                0,
                loop=connection.loop,
            )
            code, info = "227", ["listen socket created"]

        else:

            code, info = "227", ["listen socket already exists"]

        for sock in connection.passive_server.sockets:

            if sock.family == socket.AF_INET:

                host, port = sock.getsockname()
                break

        nums = tuple(map(int, str.split(host, "."))) + (port >> 8, port & 0xff)
        info.append(str.format("({})", str.join(",", map(str, nums))))

        connection.response(code, info)
        return True

    @ConnectionConditions(ConnectionConditions.login_required)
    @asyncio.coroutine
    def abor(self, connection, rest):

        connection.abort = True
        connection.response("150", "abort requested")
        return True

    @asyncio.coroutine
    def appe(self, connection, rest):

        return (yield from self.stor(connection, rest, "ab"))



from . import common


__all__ = (
    "StatusCodeError",
    "PathIsNotFileOrDir",
    "PathIsNotAbsolute",
)


class StatusCodeError(Exception):
    """
    Raised for unexpected or "bad" status codes.

    :param expected_codes: tuple of expected codes or expected code
    :type expected_codes: :py:class:`tuple` of :py:class:`aioftp.Code` or
        :py:class:`aioftp.Code`

    :param received_codes: tuple of received codes or received code
    :type received_codes: :py:class:`tuple` of :py:class:`aioftp.Code` or
        :py:class:`aioftp.Code`

    :param info: list of lines with server response
    :type info: :py:class:`list` of :py:class:`str`

    Usage::

        >>> try:
        ...     # something with aioftp
        ... except StatusCodeError as e:
        ...     print(e.expected_codes, e.received_codes, e.info)
        ...     # analyze state

    Exception members are tuples, even for one code.
    """

    def __init__(self, expected_codes, received_codes, info):

        super().__init__(
            str.format(
                "Waiting for {} but got {} {}",
                expected_codes,
                received_codes,
                repr(info),
            )
        )
        self.expected_codes = common.wrap_with_container(expected_codes)
        self.received_codes = common.wrap_with_container(received_codes)
        self.info = info


class PathIsNotFileOrDir(Exception):
    """
    Raised when "path abstract layer" tells, that `path` is not a file and
        is not a directory.
    """

    pass


class PathIsNotAbsolute(Exception):
    """
    Raised when "path" is not absolute.
    """

    pass



from common import *


@aioftp_setup(
    server_args=([(aioftp.User(
        base_path="tests/foo",
        home_path="/",
        permissions=[aioftp.Permission(writable=False)],
    ),)], {}))
@with_connection
@expect_codes_in_exception("550")
def test_permission_denied(loop, client, server):

    yield from client.login()
    yield from client.make_directory("bar")
    yield from client.quit()


@aioftp_setup(
    server_args=([(aioftp.User(
        base_path="tests/foo",
        home_path="/",
        permissions=[
            aioftp.Permission("/", writable=False),
            aioftp.Permission("/bar"),
            aioftp.Permission("/foo"),
        ],
    ),)], {}))
@with_connection
@with_tmp_dir("foo")
def test_permission_overriden(loop, client, server, *, tmp_dir):

    yield from client.login()
    yield from client.make_directory("bar")
    yield from client.remove_directory("bar")
    yield from client.quit()


def test_permission_representation():

    p = aioftp.Permission(writable=False)
    nose.tools.eq_(
        repr(p),
        "Permission(PurePosixPath('/'), readable=True, writable=False)"
    )



import logging
import collections
import time
import asyncio


__all__ = ("Throttle", "ThrottleMemory", "end_of_line", "default_block_size")


logger = logging.getLogger("aioftp")
end_of_line = "\r\n"
default_block_size = 8192


def wrap_with_container(o):

    if isinstance(o, str):

        o = (o,)

    return o


class ThrottleMemory:

    def __init__(self):

        self.end = 0

    def append(self, data, throttle):

        count = len(data)
        now = time.perf_counter()
        self.end = max(now, self.end) + count / throttle

    def timeout(self):

        return max(0, self.end - time.perf_counter())


class Throttle:

    def __init__(self, stream, *, loop, throttle=None, memory=None):

        self.stream = stream
        self.loop = loop
        self.throttle = throttle
        self.memory = memory or ThrottleMemory()

    @asyncio.coroutine
    def read(self, count=default_block_size):

        data = yield from self.stream.read(count)
        if self.throttle is not None:

            self.memory.append(data, self.throttle)
            yield from asyncio.sleep(self.memory.timeout(), loop=self.loop)

        return data

    @asyncio.coroutine
    def readline(self):

        data = yield from self.stream.readline()
        if self.throttle is not None:

            self.memory.append(data, self.throttle)
            yield from asyncio.sleep(self.memory.timeout(), loop=self.loop)

        return data

    def write(self, data):

        self.stream.write(data)
        if self.throttle is not None:

            self.memory.append(data, self.throttle)

    @asyncio.coroutine
    def drain(self):

        yield from self.stream.drain()
        if self.throttle is not None:

            yield from asyncio.sleep(self.memory.timeout(), loop=self.loop)

    def __getattr__(self, name):

        return getattr(self.stream, name)



"""Simple aioftp-based server with one user (anonymous or not).

Usage: aioftp [(<login> <password>)] [options]

Options:
    -q, --quiet             set logging level to "ERROR" instead of "INFO"
    --host=host             host for binding [default: 127.0.0.1]
    --port=port             port for binding [default: 8021]
"""
import asyncio
import logging

import docopt

import aioftp


args = docopt.docopt(__doc__, version=aioftp.__version__)
print(str.format("aioftp v{}", aioftp.__version__))

if not args["--quiet"]:

    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s %(message)s",
        datefmt="[%H:%M:%S]:",
    )

user = aioftp.User(args["<login>"], args["<password>"])
server = aioftp.Server([user])

loop = asyncio.get_event_loop()
loop.run_until_complete(server.start(args["--host"], int(args["--port"])))
try:

    loop.run_forever()

except KeyboardInterrupt:

    server.close()
    loop.run_until_complete(server.wait_closed())
    loop.close()
